# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
This file implements a DC-OPF for time series
That means that solves the OPF problem for a complete time series at once
"""

from GridCal.Engine.basic_structures import MIPSolvers
from GridCal.Engine.Core.time_series_opf_data import OpfTimeCircuit
from GridCal.Engine.Simulations.OPF.opf_templates import OpfTimeSeries
from GridCal.ThirdParty.pulp import *


def add_objective_function(Pg, Pb, LSlack, FSlack1, FSlack2,
                           cost_g, cost_b, cost_l, cost_br):
    """
    Add the objective function to the problem
    :param problem: LpProblem instance
    :param Pg: generator LpVars (ng, nt)
    :param Pb: batteries LpVars (nb, nt)
    :param LSlack: Load slack LpVars (nl, nt)
    :param FSlack1: Branch overload slack1 (m, nt)
    :param FSlack2: Branch overload slack2 (m, nt)
    :param cost_g: Cost of the generators (ng, nt)
    :param cost_b: Cost of the batteries (nb, nt)
    :param cost_l: Cost of the loss of load (nl, nt)
    :param cost_br: Cost of the overload (m, nt)
    :return: Nothing, just assign the objective function
    """

    f_obj = lpSum(cost_g * Pg)

    f_obj += lpSum(cost_b * Pb)

    f_obj += lpSum(cost_l * LSlack)

    f_obj += lpSum(cost_br * (FSlack1 + FSlack2))

    return f_obj


def set_fix_generation(problem, Pg, P_profile, enabled_for_dispatch):
    """
    Set the generatin fixed at the non dispatchabel generators
    :param problem: LP problem instance
    :param Pg: Array of generation variables
    :param P_profile: Array of fixed generation values
    :param enabled_for_dispatch: array of "enables" for dispatching generators
    :return: Nothing
    """

    idx = np.where(enabled_for_dispatch == False)[0]

    lpAddRestrictions2(problem=problem,
                       lhs=Pg[idx, :],
                       rhs=P_profile[idx, :],
                       # Fmax + FSlack2
                       name='fixed_generation',
                       op='=')


def get_power_injections(C_bus_gen, Pg, C_bus_bat, Pb, C_bus_load, PlSlack, QlSlack, Pl, Ql):
    """
    Create the power injections per bus
    :param C_bus_gen: Bus-Generators sparse connectivity matrix (n, ng)
    :param Pg: generator LpVars (ng, nt)
    :param C_bus_bat: Bus-Batteries sparse connectivity matrix (n, nb)
    :param Pb: Batteries LpVars (nb, nt)
    :param C_bus_load: Bus-Load sparse connectivity matrix (n, nl)
    :param PlSlack: Load (real) slack LpVars (nl, nt)
    :param QlSlack: Load (imag) slack LpVars (nl, nt)
    :param Pl: Load values (nl, nt)
    :return: Power injection at the buses (n, nt)
    """

    P = lpDot(C_bus_gen, Pg)

    P += lpDot(C_bus_bat, Pb)

    P -= lpDot(C_bus_load, Pl - PlSlack)

    Q = - lpDot(C_bus_load, Ql - QlSlack)

    return P, Q


def add_ac_nodal_power_balance(numerical_circuit: OpfTimeCircuit, problem: LpProblem, dvm, dva, P, Q, start_, end_):
    """
    Add the nodal power balance
    :param numerical_circuit: NumericalCircuit instance
    :param problem: LpProblem instance
    :param dva: Voltage angles LpVars (n, nt)
    :param P: Power injection at the buses LpVars (n, nt)
    :return: Nothing, the restrictions are added to the problem
    """

    # do the topological computation
    calc_inputs = numerical_circuit.split_into_islands()

    # generate the time indices to simulate
    if end_ == -1:
        end_ = len(numerical_circuit.time_array)

    nodal_restrictions_P = np.empty((numerical_circuit.nbus, end_ - start_), dtype=object)
    nodal_restrictions_Q = np.empty((numerical_circuit.nbus, end_ - start_), dtype=object)

    # For every island, run the time series
    for i, calc_inpt in enumerate(calc_inputs):

        # find the original indices
        bus_original_idx = np.array(calc_inpt.original_bus_idx)

        # re-pack the variables for the island and time interval
        P_island = P[bus_original_idx, :]  # the sizes already reflect the correct time span
        Q_island = Q[bus_original_idx, :]  # the sizes already reflect the correct time span
        dva_island = dva[bus_original_idx, :]  # the sizes already reflect the correct time span
        dvm_island = dvm[bus_original_idx, :]  # the sizes already reflect the correct time span
        B_island = calc_inpt.Ybus.imag
        G_island = calc_inpt.Ybus.real
        Bs_island = calc_inpt.Yseries.imag
        Gs_island = calc_inpt.Yseries.real

        pqpv = calc_inpt.pqpv
        pq = calc_inpt.pq
        pv = calc_inpt.pv
        vd = calc_inpt.vd
        vdpv = np.r_[vd, pv]
        vdpv.sort()

        # Add nodal real power balance for the non slack nodes
        idx = bus_original_idx[pqpv]
        nodal_restrictions_P[idx] = lpAddRestrictions2(problem=problem,
                                                       lhs=-lpDot(Bs_island[np.ix_(pqpv, pqpv)], dva_island[pqpv, :])
                                                           + lpDot(G_island[np.ix_(pqpv, pq)], dvm_island[pq, :]),
                                                       rhs=P_island[pqpv, :],
                                                       name='Nodal_real_power_balance_pqpv_is' + str(i),
                                                       op='=')

        # Add nodal reactive power balance for the non slack nodes
        idx = bus_original_idx[pq]
        nodal_restrictions_Q[idx] = lpAddRestrictions2(problem=problem,
                                                       lhs=-lpDot(Gs_island[np.ix_(pq, pqpv)], dva_island[pqpv, :])
                                                           - lpDot(B_island[np.ix_(pq, pq)], dvm_island[pq, :]),
                                                       rhs=Q_island[pq, :],
                                                       name='Nodal_imag_power_balance_pqpv_is' + str(i),
                                                       op='=')

        # Add nodal real power balance for the slack nodes
        idx = bus_original_idx[vd]
        nodal_restrictions_P[idx] = lpAddRestrictions2(problem=problem,
                                                       lhs=-lpDot(Bs_island[vd, :], dva_island)
                                                           + lpDot(G_island[vd, :], dvm_island),
                                                       rhs=P_island[vd, :],
                                                       name='Nodal_real_power_balance_vd_is' + str(i),
                                                       op='=')

        # delta of voltage angles equal to zero for the slack nodes (vd)
        lpAddRestrictions2(problem=problem,
                           lhs=dva_island[vd, :],
                           rhs=np.zeros_like(dva_island[vd, :]),
                           name='dVa_vd_zero_is' + str(i),
                           op='=')

        # delta of voltage module equal to zero for the slack and pv nodes (vdpv)
        lpAddRestrictions2(problem=problem,
                           lhs=dvm_island[vdpv, :],
                           rhs=np.zeros_like(dva_island[vdpv, :]),
                           name='dVm_vdpv_zero_is' + str(i),
                           op='=')

    return nodal_restrictions_P, nodal_restrictions_Q


def add_branch_loading_restriction(problem: LpProblem,
                                   theta_f, theta_t, Bseries,
                                   Fmax, FSlack1, FSlack2):
    """
    Add the branch loading restrictions
    :param problem: LpProblem instance
    :param theta_f: voltage angles at the "from" side of the branches (m, nt)
    :param theta_t: voltage angles at the "to" side of the branches (m, nt)
    :param Bseries: Array of branch susceptances (m)
    :param Fmax: Array of branch ratings (m, nt)
    :param FSlack1: Array of branch loading slack variables in the from-to sense
    :param FSlack2: Array of branch loading slack variables in the to-from sense
    :return: Nothing
    """

    load_f = Bseries * (theta_f - theta_t)
    load_t = Bseries * (theta_t - theta_f)

    # from-to branch power restriction
    lpAddRestrictions2(problem=problem,
                       lhs=load_f,
                       rhs=np.array([Fmax[:, i] + FSlack1[:, i] for i in range(FSlack1.shape[1])]).transpose(),  # Fmax + FSlack1
                       name='from_to_branch_rate',
                       op='<=')

    # to-from branch power restriction
    lpAddRestrictions2(problem=problem,
                       lhs=load_t,
                       rhs=np.array([Fmax[:, i] + FSlack2[:, i] for i in range(FSlack2.shape[1])]).transpose(),  # Fmax + FSlack2
                       name='to_from_branch_rate',
                       op='<=')

    return load_f, load_t


def add_battery_discharge_restriction(problem: LpProblem, SoC0, Capacity, Efficiency, Pb, E, dt):
    """
    Add the batteries capacity restrictions
    :param problem: LpProblem instance
    :param SoC0: State of Charge at 0 (nb)
    :param SoCmax: Maximum State of Charge (nb)
    :param SoCmin: Minimum State of Charge (nb)
    :param Capacity: Capacities of the batteries (nb) in MWh/MW base
    :param Efficiency: Roundtrip efficiency
    :param Pb: Batteries injection power LpVars (nb, nt)
    :param E: Batteries Energy state LpVars (nb, nt)
    :param dt: time increments in hours (nt-1)
    :return: Nothing, the restrictions are added to the problem
    """

    # set the initial state of charge
    lpAddRestrictions2(problem=problem,
                       lhs=E[:, 0],
                       rhs=SoC0 * Capacity,
                       name='initial_soc',
                       op='=')

    # compute the inverse of he efficiency because pulp does not divide by floats
    eff_inv = 1 / Efficiency

    # set the Energy values for t=1:nt
    for i in range(len(dt) - 1):

        t = i + 1

        # set the energy value Et = E(t-1) + dt * Pb / eff
        lpAddRestrictions2(problem=problem,
                           lhs=E[:, t],
                           rhs=E[:, t-1] - dt[i] * Pb[:, t] * eff_inv,
                           name='initial_soc_t' + str(t) + '_',
                           op='=')


class OpfAcTimeSeries(OpfTimeSeries):

    def __init__(self, numerical_circuit: OpfTimeCircuit, start_idx, end_idx, solver_type: MIPSolvers = MIPSolvers.CBC):
        """
        AC time series linear optimal power flow
        :param numerical_circuit: NumericalCircuit instance
        :param start_idx: start index of the time series
        :param end_idx: end index of the time series
        :param solver_type: MIP solver_type to use
        :param batteries_energy_0: initial state of the batteries, if None the default values are taken
        """

        OpfTimeSeries.__init__(self, numerical_circuit=numerical_circuit, start_idx=start_idx, end_idx=end_idx,
                               solver_type=solver_type, skip_formulation=True)

        self.v0 = None
        self.dva = None
        self.dvm = None
        self.Ql = None

    def formulate(self, batteries_energy_0=None):
        """
        Formulate the AC OPF time series in the non-sequential fashion (all to the solver_type at once)
        :param batteries_energy_0: initial energy state of the batteries (if none, the default is taken)
        :return: PuLP Problem instance
        """
        numerical_circuit = self.numerical_circuit

        # general indices
        n = numerical_circuit.nbus
        m = numerical_circuit.nbr
        ng = numerical_circuit.ngen
        nb = numerical_circuit.nbatt
        nl = numerical_circuit.nload
        nt = self.end_idx - self.start_idx
        a = self.start_idx
        b = self.end_idx
        Sbase = numerical_circuit.Sbase

        # battery
        Capacity = numerical_circuit.battery_enom / Sbase
        minSoC = numerical_circuit.battery_min_soc
        maxSoC = numerical_circuit.battery_max_soc

        if batteries_energy_0 is None:
            SoC0 = numerical_circuit.battery_soc_0
        else:
            SoC0 = (batteries_energy_0 / Sbase) / Capacity

        Pb_max = numerical_circuit.battery_pmax / Sbase
        Pb_min = numerical_circuit.battery_pmin / Sbase
        Efficiency = (numerical_circuit.battery_discharge_efficiency + numerical_circuit.battery_charge_efficiency) / 2.0
        cost_b = numerical_circuit.battery_cost[:, a:b]

        # generator
        Pg_max = numerical_circuit.generator_pmax / Sbase
        Pg_min = numerical_circuit.generator_pmin / Sbase
        P_profile = numerical_circuit.generator_p[:, a:b] / Sbase
        cost_g = numerical_circuit.generator_cost[:, a:b]
        enabled_for_dispatch = numerical_circuit.generator_dispatchable

        # load
        Pl = (numerical_circuit.load_active[:, a:b] * numerical_circuit.load_s.real[:, a:b]) / Sbase
        Ql = (numerical_circuit.load_active[:, a:b] * numerical_circuit.load_s.imag[:, a:b]) / Sbase
        cost_l = numerical_circuit.load_cost[:, a:b]

        # branch
        branch_ratings = numerical_circuit.branch_rates[:, a:b] / Sbase
        ys = 1 / (numerical_circuit.branch_R + 1j * numerical_circuit.branch_X)
        Bseries = (numerical_circuit.branch_active[:, a:b].T * ys.imag).T
        cost_br = numerical_circuit.branch_cost[:, a:b]

        # Compute time delta in hours
        dt = np.zeros(nt)  # here nt = end_idx - start_idx
        for t in range(1, nt):
            dt[t - 1] = (numerical_circuit.time_array[a + t] - numerical_circuit.time_array[a + t - 1]).seconds / 3600

        # create LP variables
        Pg = lpMakeVars(name='Pg', shape=(ng, nt), lower=Pg_min, upper=Pg_max)
        Pb = lpMakeVars(name='Pb', shape=(nb, nt), lower=Pb_min, upper=Pb_max)
        E = lpMakeVars(name='E', shape=(nb, nt), lower=Capacity * minSoC, upper=Capacity * maxSoC)
        load_slack = lpMakeVars(name='LSlack', shape=(nl, nt), lower=0, upper=None)
        dva = lpMakeVars(name='dva', shape=(n, nt), lower=-3.14, upper=3.14)
        dvm = lpMakeVars(name='dvm', shape=(n, nt), lower=0, upper=2)
        theta_f = dva[numerical_circuit.F, :]
        theta_t = dva[numerical_circuit.T, :]
        branch_rating_slack1 = lpMakeVars(name='FSlack1', shape=(m, nt), lower=0, upper=None)
        branch_rating_slack2 = lpMakeVars(name='FSlack2', shape=(m, nt), lower=0, upper=None)

        # declare problem
        problem = LpProblem(name='AC_OPF_Time_Series')

        # add the objective function
        problem += add_objective_function(Pg, Pb, load_slack, branch_rating_slack1, branch_rating_slack2,
                                          cost_g, cost_b, cost_l, cost_br)

        # set the fixed generation values
        set_fix_generation(problem=problem, Pg=Pg, P_profile=P_profile,
                           enabled_for_dispatch=enabled_for_dispatch)

        # compute the power injections per node
        P, Q = get_power_injections(C_bus_gen=numerical_circuit.generator_data.C_bus_gen, Pg=Pg,
                                    C_bus_bat=numerical_circuit.battery_data.C_bus_batt, Pb=Pb,
                                    C_bus_load=numerical_circuit.load_data.C_bus_load,
                                    PlSlack=load_slack, QlSlack=load_slack,
                                    Pl=Pl, Ql=Ql)

        # compute the nodal power balance restrictions
        nodal_restrictions_P, nodal_restrictions_Q = add_ac_nodal_power_balance(numerical_circuit=numerical_circuit,
                                                                                problem=problem,
                                                                                dvm=dvm, dva=dva, P=P, Q=Q,
                                                                                start_=self.start_idx, end_=self.end_idx)

        load_f, load_t = add_branch_loading_restriction(problem, theta_f, theta_t, Bseries, branch_ratings,
                                                        branch_rating_slack1, branch_rating_slack2)

        # if there are batteries, add the batteries
        if nb > 0:
            add_battery_discharge_restriction(problem, SoC0, Capacity, Efficiency, Pb, E, dt)

        # Assign variables to keep
        # transpose them to be in the format of GridCal: time, device
        self.v0 = np.abs(numerical_circuit.Vbus[:, a:b]).T
        self.dva = dva.transpose()
        self.dvm = dvm.transpose()
        self.Pg = Pg.transpose()
        self.Pb = Pb.transpose()
        self.Pl = Pl.transpose()
        self.E = E.transpose()
        self.load_shedding = load_slack.transpose()
        self.s_from = load_f.transpose()
        self.s_to = load_t.transpose()
        self.overloads = (branch_rating_slack1 + branch_rating_slack2).transpose()
        self.rating = branch_ratings.T
        self.nodal_restrictions = nodal_restrictions_P

        return problem

    def get_voltage(self):
        """
        return the complex voltages (time, device)
        :return: 2D array
        """
        angles = self.extract2D(self.dva)
        modules = self.v0 + self.extract2D(self.dvm)
        return modules * np.exp(-1j * angles)


if __name__ == '__main__':

    from GridCal.Engine import *

    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/Lynn 5 Bus pv.gridcal'
    fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE39_1W.gridcal'
    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/grid_2_islands.xlsx'

    main_circuit = FileOpen(fname).open()

    # main_circuit.buses[3].controlled_generators[0].enabled_dispatch = False

    # get the power flow options from the GUI
    solver = SolverType.AC_OPF
    mip_solver = MIPSolvers.CBC
    grouping = TimeGrouping.Daily
    pf_options = PowerFlowOptions()

    options = OptimalPowerFlowOptions(solver=solver,
                                      time_grouping=grouping,
                                      mip_solver=mip_solver,
                                      power_flow_options=pf_options)

    start = 0
    end = len(main_circuit.time_profile)

    # create the OPF time series instance
    # if non_sequential:
    optimal_power_flow_time_series = OptimalPowerFlowTimeSeries(grid=main_circuit,
                                                                options=options,
                                                                start_=start,
                                                                end_=end)

    optimal_power_flow_time_series.run()

    v = optimal_power_flow_time_series.results.voltage
    print('Angles\n', np.angle(v))

    l = optimal_power_flow_time_series.results.loading
    print('Branch loading\n', l)

    g = optimal_power_flow_time_series.results.generator_power
    print('Gen power\n', g)

    pr = optimal_power_flow_time_series.results.shadow_prices
    print('Nodal prices \n', pr)


# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import json

import numpy as np
import numba as nb
from GridCal.Engine.basic_structures import Logger
from GridCal.Engine.Core.multi_circuit import MultiCircuit
from GridCal.Engine.Devices import *


@nb.jit()
def compress_array_numba(value, base):
    data = list()
    indptr = list()
    for i, x in enumerate(value):
        if x != base:
            data.append(x)
            indptr.append(i)
    return data, indptr


def compress_array(arr, min_sparsity=0.2):
    """
    Compress array
    :param arr: list of np.ndarray
    :param min_sparsity: minimum sparsity of the array to consider it sparse
    :return: dictionary with at least type and data entries.
            if sparse: {'type': 'sparse',
                        'base': base,
                        'size': len(arr),
                        'data': data,
                        'indptr': indptr}

            if dense: {'type': 'dense',
                       'data': arr}
    """
    if isinstance(arr, list) or isinstance(arr, np.ndarray):
        u = np.unique(arr)
        f = len(u) / len(arr)  # sparsity factor
        if f < min_sparsity:
            base = u[0]  # pick the first always as the base
            if isinstance(base, np.bool_):
                base = bool(base)
            data = list()
            indptr = list()
            if len(u) > 1:
                if isinstance(arr, list):
                    data, indptr = compress_array_numba(nb.typed.List(arr), base)
                elif isinstance(arr, np.ndarray):
                    data, indptr = compress_array_numba(arr, base)
                else:
                    raise Exception('Unknown profile type' + str(type(arr)))

            return {'type': 'sparse',
                    'base': base,
                    'size': len(arr),
                    'data': data,
                    'indptr': indptr}
        else:
            return {'type': 'dense',
                    'data': arr}
    else:
        raise Exception('Unknown profile type' + str(type(arr)))


def decompress_array(d):
    """
    decompress array (in profile form or list form)
    :param d: dictionary containing a profile or a simple list of values
    :return: numpy array
    """
    if isinstance(d, dict):
        if 'type' in d.keys():
            if d['type'] == 'sparse':
                n = d['size']
                val = np.full(n, d['base'], dtype=type(d['base']))
                for i, x in zip(d['indptr'], d['data']):
                    val[i] = x
                return val

            elif d['type'] == 'dense':
                val = np.array(d['data'])
                return val
            else:
                raise Exception('Unknown profile type' + str(d['type']))

        else:
            raise Exception("The passed dictionary is not a profile definition")
    elif isinstance(d, list):
        return np.array(d)
    else:
        raise Exception("The passed value is not a list or dictionary definition")


def convert_to_sparse(d: dict, min_sparsity=0.2):
    """
    Convert a dictionary of profiles to a dictionary of sparse or dense profiles
    :param d: dictionary of profiles i.e. {"p": [1.2, 3.2, ...], "q": [0.3, 1.1, ...]}
    :param min_sparsity: minimum sparsity of the array to consider it sparse
    :return: dictionary of profiles but the profiles are objects that indicate if the profile is sparse or dense
    """
    for key, value in d.items():
        if isinstance(value, list) or isinstance(value, np.ndarray):
            d[key] = compress_array(value, min_sparsity)

    return d


def parse_json_data(data) -> MultiCircuit:
    """
    Parse JSON structure into GridCal MultiCircuit
    :param data: JSON structure (list of dictionaries)
    :return: GridCal MultiCircuit
    """

    circuit = MultiCircuit()

    bus_id = dict()

    for element in data:

        if element["phases"] == 'ps':

            if element["type"] == "circuit":

                circuit = MultiCircuit()
                circuit.name = element["name"]
                circuit.Sbase = element["Sbase"]
                circuit.comments = element['comments']

            elif element["type"] == "bus":

                # create the bus and add some properties
                elm = Bus(name=element["name"],
                          vnom=element["Vnom"],
                          vmin=0.9,
                          vmax=1.1,
                          xpos=element['x'],
                          ypos=element['y'],
                          height=element['h'],
                          width=element['w'],
                          active=True)

                if element["is_slack"]:
                    elm.type = BusMode.Slack
                if element["vmax"] > 0:
                    elm.Vmax = element["vmax"]
                if element["vmin"] > 0:
                    elm.Vmin = element["vmin"]

                elm.Zf = complex(element['rf'], element['xf'])

                circuit.add_bus(elm)

                # add the bus to the dictionary
                bus_id[element["id"]] = elm

            elif element["type"] == "load":

                # get the matching bus object pointer
                bus = bus_id[element["bus"]]

                # create a load in the bus
                elm = Load(name=element['name'],
                           G=element["G"],
                           B=element["B"],
                           Ir=element["Ir"],
                           Ii=element["Ii"],
                           P=element["P"],
                           Q=element["Q"],
                           active=element['active'])
                bus.loads.append(elm)

            elif element["type"] == "controlled_gen":

                # get the matching bus object pointer
                bus = bus_id[element["bus"]]

                # create a load in the bus
                elm = Generator(name=element['name'],
                                active_power=element["P"],
                                voltage_module=element["vset"],
                                Qmin=element['qmin'],
                                Qmax=element['qmax'],
                                Snom=element['Snom'],
                                power_prof=None,
                                vset_prof=None,
                                active=element['active'],
                                p_min=0.0,
                                p_max=element['Snom'],
                                op_cost=1.0)
                bus.controlled_generators.append(elm)

            elif element["type"] == "static_gen":

                # get the matching bus object pointer
                bus = bus_id[element["bus"]]

                # create a load in the bus
                elm = StaticGenerator(name=element['name'],
                                      P=element['P'], Q=element['Q'],
                                      active=element['active'])
                bus.static_generators.append(elm)

            elif element["type"] == "battery":

                # get the matching bus object pointer
                bus = bus_id[element["bus"]]

                # create a load in the bus
                elm = Battery(name=element['name'],
                              active_power=element["P"],
                              voltage_module=element["vset"],
                              Qmin=element['qmin'],
                              Qmax=element['qmax'],
                              Snom=element['Snom'],
                              Enom=element['Enom'],
                              power_prof=None,
                              vset_prof=None,
                              active=element['active'])
                bus.batteries.append(elm)

            elif element["type"] == "shunt":

                # get the matching bus object pointer
                bus = bus_id[element["bus"]]

                # create a load in the bus
                elm = Shunt(name=element['name'],
                            G=element["g"], B=element["b"],
                            active=element['active'])
                bus.shunts.append(elm)

            elif element["type"] == "branch":

                # get the matching bus object pointer
                bus1 = bus_id[element["from"]]
                bus2 = bus_id[element["to"]]

                # create a load in the  bus
                elm = Branch(bus_from=bus1,
                             bus_to=bus2,
                             name=element["name"],
                             r=element["r"],
                             x=element["x"],
                             g=element["g"],
                             b=element["b"],
                             rate=element["rate"],
                             tap=element["tap_module"],
                             shift_angle=element["tap_angle"],
                             active=element["active"],
                             mttf=0,
                             mttr=0,
                             branch_type=element["branch_type"])
                circuit.add_branch(elm)

        else:
            warn('ID: ' + element["id"] + ' error: GridCal only takes positive sequence elements.')

    return circuit


def set_object_properties(elm, prop: list, entry: dict):
    for jprop, gcprop in prop:
        if jprop in entry.keys():
            setattr(elm, gcprop, entry[jprop])


def parse_json_data_v3(data: dict, logger: Logger):
    """
    Json parser for V3
    :param data:
    :param logger:
    :return:
    """
    devices = data['devices']
    profiles = data['profiles']

    # parse devices
    if DeviceType.CircuitDevice.value in devices.keys():

        dta = devices[DeviceType.CircuitDevice.value]

        circuit = MultiCircuit(name=str(dta['name']),
                               Sbase=float(dta['sbase']),
                               fbase=float(dta['fbase']),
                               idtag=str(dta['id']))

        jcircuit = devices["Circuit"]
        circuit.Sbase = jcircuit["sbase"]

        # parse time series
        if DeviceType.CircuitDevice.value in profiles.keys():
            circuit.set_time_profile(profiles[DeviceType.CircuitDevice.value]['time'])

        # Countries
        country_dict = dict()
        if 'Country' in devices.keys():
            elms = devices["Country"]
            for jentry in elms:
                elm = Country(idtag=str(jentry['id']),
                              code=str(jentry['code']),
                              name=str(jentry['name']))

                circuit.countries.append(elm)
                country_dict[elm.idtag] = elm

        else:
            elm = Country(idtag=None, code='Default', name='Default')
            circuit.countries.append(elm)

        # Areas
        areas_dict = dict()
        if 'Area' in devices.keys():
            elms = devices["Area"]
            for jentry in elms:
                elm = Area(idtag=str(jentry['id']),
                           code=str(jentry['code']),
                           name=str(jentry['name']))
                circuit.areas.append(elm)
                areas_dict[elm.idtag] = elm
        else:
            elm = Area(idtag=None, code='Default', name='Default')
            circuit.areas.append(elm)

        # Zones
        zones_dict = dict()
        if 'Zone' in devices.keys():
            elms = devices["Zone"]
            for jentry in elms:
                elm = Zone(idtag=str(jentry['id']),
                           code=str(jentry['code']),
                           name=str(jentry['name']))
                circuit.zones.append(elm)
                zones_dict[elm.idtag] = elm
        else:
            elm = Zone(idtag=None, code='Default', name='Default')
            circuit.zones.append(elm)

        # Substations
        substations_dict = dict()
        if 'Substation' in devices.keys():
            elms = devices["Substation"]
            for jentry in elms:
                elm = Substation(idtag=str(jentry['id']),
                                 code=str(jentry['code']),
                                 name=str(jentry['name']))
                circuit.substations.append(elm)
                substations_dict[elm.idtag] = elm
        else:
            elm = Substation(idtag=None, code='Default', name='Default')
            circuit.substations.append(elm)

        # buses
        bus_dict = dict()
        if 'Bus' in devices.keys():
            buses = devices["Bus"]

            if 'Bus' in profiles.keys():
                device_profiles_dict = {e['id']: e for e in profiles["Bus"]}
                has_profiles = True
            else:
                device_profiles_dict = dict()
                has_profiles = False

            for jentry in buses:

                area_id = str(jentry['area']) if 'area' in jentry.keys() else ''
                zone_id = str(jentry['zone']) if 'zone' in jentry.keys() else ''
                substation_id = str(jentry['substation']) if 'substation' in jentry.keys() else ''
                country_id = str(jentry['country']) if 'country' in jentry.keys() else ''

                if area_id in areas_dict.keys():
                    area = areas_dict[area_id]
                else:
                    area = circuit.areas[0]

                if zone_id in zones_dict.keys():
                    zone = zones_dict[zone_id]
                else:
                    zone = circuit.zones[0]

                if substation_id in substations_dict.keys():
                    substation = substations_dict[substation_id]
                else:
                    substation = circuit.substations[0]

                if country_id in country_dict.keys():
                    country = country_dict[country_id]
                else:
                    country = circuit.countries[0]

                bus = Bus(name=str(jentry['name']),
                          idtag=str(jentry['id']),
                          code=str(jentry['name_code']),
                          vnom=float(jentry['vnom']),
                          vmin=float(jentry['vmin']),
                          vmax=float(jentry['vmax']),
                          r_fault=float(jentry['rf']),
                          x_fault=float(jentry['xf']),
                          xpos=float(jentry['x']),
                          ypos=float(jentry['y']),
                          height=float(jentry['h']),
                          width=float(jentry['w']),
                          active=bool(jentry['active']),
                          is_slack=bool(jentry['is_slack']),
                          area=area,
                          zone=zone,
                          substation=substation,
                          country=country,
                          longitude=float(jentry['lon']),
                          latitude=float(jentry['lat']))

                bus_dict[jentry['id']] = bus

                if has_profiles:
                    bus.active_prof = decompress_array(device_profiles_dict[bus.idtag]['active'])

                circuit.add_bus(bus)

        if 'Generator' in devices.keys():
            generators = devices["Generator"]

            if 'Generator' in profiles.keys():
                device_profiles_dict = {e['id']: e for e in profiles["Generator"]}
                has_profiles = True
            else:
                device_profiles_dict = dict()
                has_profiles = False

            for jentry in generators:
                elm = Generator(name=str(jentry['name']),
                                idtag=str(jentry['id']),
                                active_power=float(jentry['p']),
                                power_factor=float(jentry['pf']),
                                voltage_module=float(jentry['vset']),
                                is_controlled=bool(jentry['is_controlled']),
                                Qmin=float(jentry['qmin']),
                                Qmax=float(jentry['qmax']),
                                Snom=float(jentry['snom']),
                                active=bool(jentry['active']),
                                p_min=float(jentry['pmin']),
                                p_max=float(jentry['pmax']),
                                op_cost=float(jentry['cost']),
                                )

                if has_profiles:
                    profile_entry = device_profiles_dict[elm.idtag]
                    elm.active_prof = decompress_array(profile_entry['active'])
                    elm.P_prof = decompress_array(profile_entry['p'])
                    elm.Vset_prof = decompress_array(profile_entry['v'])
                    elm.Pf_prof = decompress_array(profile_entry['pf'])

                elm.bus = bus_dict[jentry['bus']]
                circuit.add_generator(elm.bus, elm)

        if 'Battery' in devices.keys():
            batteries = devices["Battery"]

            if 'Battery' in profiles.keys():
                device_profiles_dict = {e['id']: e for e in profiles["Battery"]}
                has_profiles = True
            else:
                device_profiles_dict = dict()
                has_profiles = False

            for jentry in batteries:
                elm = Battery(name=str(jentry['name']),
                              idtag=str(jentry['id']),
                              active_power=float(jentry['p']),
                              power_factor=float(jentry['pf']),
                              voltage_module=float(jentry['vset']),
                              is_controlled=bool(jentry['is_controlled']),
                              Qmin=float(jentry['qmin']),
                              Qmax=float(jentry['qmax']),
                              Snom=float(jentry['snom']),
                              active=bool(jentry['active']),
                              p_min=float(jentry['pmin']),
                              p_max=float(jentry['pmax']),
                              op_cost=float(jentry['cost']),
                              )

                if has_profiles:
                    profile_entry = device_profiles_dict[elm.idtag]
                    elm.active_prof = decompress_array(profile_entry['active'])
                    elm.P_prof = decompress_array(profile_entry['p'])
                    elm.Vset_prof = decompress_array(profile_entry['v'])
                    elm.Pf_prof = decompress_array(profile_entry['pf'])

                elm.bus = bus_dict[jentry['bus']]
                circuit.add_battery(elm.bus, elm)

        if 'Load' in devices.keys():
            loads = devices["Load"]

            if 'Load' in profiles.keys():
                device_profiles_dict = {e['id']: e for e in profiles["Load"]}
                has_profiles = True
            else:
                device_profiles_dict = dict()
                has_profiles = False

            for jentry in loads:
                elm = Load(name=str(jentry['name']),
                           idtag=str(jentry['id']),
                           P=float(jentry['p']),
                           Q=float(jentry['q']),
                           Ir=float(jentry['ir']),
                           Ii=float(jentry['ii']),
                           G=float(jentry['g']),
                           B=float(jentry['b']),
                           active=bool(jentry['active']))

                if has_profiles:
                    profile_entry = device_profiles_dict[elm.idtag]
                    elm.active_prof = decompress_array(profile_entry['active'])
                    elm.P_prof = decompress_array(profile_entry['p'])
                    elm.Q_prof = decompress_array(profile_entry['q'])
                    elm.Ir_prof = decompress_array(profile_entry['ir'])
                    elm.Ii_prof = decompress_array(profile_entry['ii'])
                    elm.G_prof = decompress_array(profile_entry['g'])
                    elm.B_prof = decompress_array(profile_entry['b'])

                elm.bus = bus_dict[jentry['bus']]
                circuit.add_load(elm.bus, elm)

        if "Shunt" in devices.keys():
            shunts = devices["Shunt"]

            if 'Shunt' in profiles.keys():
                device_profiles_dict = {e['id']: e for e in profiles["Shunt"]}
                has_profiles = True
            else:
                device_profiles_dict = dict()
                has_profiles = False

            for jentry in shunts:
                if 'b_max' in jentry:
                    Bmax = float(jentry['b_max'])
                elif 'bmax' in jentry:
                    Bmax = float(jentry['bmax'])
                else:
                    Bmax = 9999

                if 'b_min' in jentry:
                    Bmin = float(jentry['b_min'])
                elif 'bmax' in jentry:
                    Bmin = float(jentry['bmin'])
                else:
                    Bmin = 9999

                elm = Shunt(name=str(jentry['name']),
                            idtag=str(jentry['id']),
                            G=float(jentry['g']),
                            B=float(jentry['b']),
                            Bmax=Bmax,
                            Bmin=Bmin,
                            active=bool(jentry['active']),
                            controlled=bool(jentry['controlled'])
                            )

                if has_profiles:
                    profile_entry = device_profiles_dict[elm.idtag]
                    elm.active_prof = decompress_array(profile_entry['active'])
                    elm.G_prof = decompress_array(profile_entry['g'])
                    elm.B_prof = decompress_array(profile_entry['b'])

                elm.bus = bus_dict[jentry['bus']]
                circuit.add_shunt(elm.bus, elm)

        if "Line" in devices.keys():

            if 'Line' in profiles.keys():
                device_profiles_dict = {e['id']: e for e in profiles["Line"]}
                has_profiles = True
            else:
                device_profiles_dict = dict()
                has_profiles = False

            for entry in devices["Line"]:
                elm = Line(bus_from=bus_dict[entry['bus_from']],
                           bus_to=bus_dict[entry['bus_to']],
                           name=str(entry['name']),
                           idtag=str(entry['id']),
                           code=str(entry['name_code']),
                           r=float(entry['r']),
                           x=float(entry['x']),
                           b=float(entry['b']),
                           rate=float(entry['rate']),
                           active=entry['active'],
                           length=float(entry['length']),
                           temp_base=float(entry['base_temperature']),
                           temp_oper=float(entry['operational_temperature']),
                           alpha=float(entry['alpha'])
                           )

                if has_profiles:
                    profile_entry = device_profiles_dict[elm.idtag]
                    elm.active_prof = decompress_array(profile_entry['active'])
                    elm.rate_prof = decompress_array(profile_entry['rate'])

                circuit.add_line(elm)

        if "Transformer2w" in devices.keys() or "Transformer" in devices.keys():

            if "Transformer2w" in devices.keys():
                transformers = devices["Transformer2w"]
                if 'Transformer2w' in profiles.keys():
                    device_profiles_dict = {e['id']: e for e in profiles["Transformer2w"]}
                    has_profiles = True
                else:
                    device_profiles_dict = dict()
                    has_profiles = False
            else:
                transformers = devices["Transformer"]
                if 'Transformer' in profiles.keys():
                    device_profiles_dict = {e['id']: e for e in profiles["Transformer"]}
                    has_profiles = True
                else:
                    device_profiles_dict = dict()
                    has_profiles = False

            for entry in transformers:

                v1 = float(entry['Vnomf'])
                v2 = float(entry['Vnomt'])

                bus_from = bus_dict[entry['bus_from']]
                bus_to = bus_dict[entry['bus_to']]

                if v1 == 0.0:
                    v1 = bus_from.Vnom

                if v2 == 0.0:
                    v2 = bus_to.Vnom

                if v1 > v2:
                    HV = v1
                    LV = v2
                else:
                    HV = v2
                    LV = v1

                elm = Transformer2W(bus_from=bus_from,
                                    bus_to=bus_to,
                                    name=str(entry['name']),
                                    idtag=str(entry['id']),
                                    code=str(entry['name_code']),
                                    r=float(entry['r']),
                                    x=float(entry['x']),
                                    g=float(entry['g']),
                                    b=float(entry['b']),
                                    rate=float(entry['rate']),
                                    HV=HV,
                                    LV=LV,
                                    active=bool(entry['active']),
                                    tap=float(entry['tap_module']),
                                    shift_angle=float(entry['tap_angle']),
                                    bus_to_regulated=bool(entry['bus_to_regulated']) if 'bus_to_regulated' in entry else False,
                                    vset=float(entry['vset']),
                                    temp_base=float(entry['base_temperature']),
                                    temp_oper=float(entry['operational_temperature']),
                                    alpha=float(entry['alpha'])
                                    )

                if has_profiles:
                    profile_entry = device_profiles_dict[elm.idtag]
                    elm.active_prof = decompress_array(profile_entry['active'])
                    elm.rate_prof = decompress_array(profile_entry['rate'])

                circuit.add_transformer2w(elm)

        if "TransformerNw" in devices.keys():

            for tr_entry in devices["TransformerNw"]:
                for entry in tr_entry['windings']:
                    v1 = float(entry['Vnomf'])
                    v2 = float(entry['Vnomt'])

                    if v1 > v2:
                        HV = v1
                        LV = v2
                    else:
                        HV = v2
                        LV = v1

                    elm = Transformer2W(bus_from=bus_dict[entry['bus_from']],
                                        bus_to=bus_dict[entry['bus_to']],
                                        name=str(tr_entry['name']),
                                        idtag=str(entry['id']),
                                        code=str(entry['name_code']),
                                        active=bool(tr_entry['active']),
                                        r=float(entry['r']),
                                        x=float(entry['x']),
                                        g=float(entry['g']),
                                        b=float(entry['b']),
                                        rate=float(entry['rate']),
                                        HV=HV,
                                        LV=LV
                                        )
                    circuit.add_transformer2w(elm)

        if "VSC" in devices.keys():

            if 'VSC' in profiles.keys():
                device_profiles_dict = {e['id']: e for e in profiles["VSC"]}
                has_profiles = True
            else:
                device_profiles_dict = dict()
                has_profiles = False

            # json property -> gc property name
            prop = [('id', 'idtag'),
                    ('name', 'name'),
                    ('name_code', 'code'),
                    ('active', 'active'),
                    ('rate', 'rate'),
                    ('contingency_factor1', 'contingency_factor'),
                    ('r', 'R1'),
                    ('x', 'X1'),
                    ('g', 'G0'),
                    ('m', 'm'),
                    ('m_min', 'm_min'),
                    ('m_max', 'm_max'),
                    ('theta', 'theta'),
                    ('theta_min', 'theta_min'),
                    ('theta_max', 'theta_max'),
                    ('Beq', 'Beq'),
                    ('Beq_min', 'Beq_min'),
                    ('Beq_max', 'Beq_max'),
                    ('alpha1', 'alpha1'),
                    ('alpha2', 'alpha2'),
                    ('alpha3', 'alpha3'),
                    ('k', 'k'),
                    ('kdp', 'kdp'),
                    ('Pfset', 'Pdc_set'),
                    ('Qfset', 'Qac_set'),
                    ('vac_set', 'Vac_set'),
                    ('vdc_set', 'Vdc_set'),
                    ]

            modes = {0: ConverterControlType.type_0_free,
                     1: ConverterControlType.type_I_1,
                     2: ConverterControlType.type_I_2,
                     3: ConverterControlType.type_I_3,
                     4: ConverterControlType.type_II_4,
                     5: ConverterControlType.type_II_5,
                     6: ConverterControlType.type_III_6,
                     7: ConverterControlType.type_III_7,
                     8: ConverterControlType.type_IV_I,
                     9: ConverterControlType.type_IV_II}

            for entry in devices["VSC"]:

                elm = VSC()
                elm.bus_from = bus_dict[entry['bus_from']]
                elm.bus_to = bus_dict[entry['bus_to']]
                set_object_properties(elm, prop, entry)

                if "control_mode" in entry.keys():
                    elm.control_mode = modes[entry["control_mode"]]
                elif "mode" in entry.keys():
                    elm.control_mode = modes[entry["mode"]]

                if has_profiles:
                    profile_entry = device_profiles_dict[elm.idtag]
                    elm.active_prof = decompress_array(profile_entry['active'])
                    elm.rate_prof = decompress_array(profile_entry['rate'])

                circuit.add_vsc(elm)

        if "HVDC Line" in devices.keys():

            if 'HVDC Line' in profiles.keys():
                device_profiles_dict = {e['id']: e for e in profiles["HVDC Line"]}
                has_profiles = True
            else:
                device_profiles_dict = dict()
                has_profiles = False

            hvdc_ctrl_dict = dict()
            hvdc_ctrl_dict[HvdcControlType.type_1_Pset.value] = HvdcControlType.type_1_Pset
            hvdc_ctrl_dict[HvdcControlType.type_0_free.value] = HvdcControlType.type_0_free

            prop = [('id', 'idtag'),
                    ('name', 'name'),
                    ('name_code', 'code'),
                    ('active', 'active'),
                    ('rate', 'rate'),
                    ('Pset', 'Pset'),
                    ('vset_from', 'Vset_f'),
                    ('vset_to', 'Vset_t'),
                    ('contingency_factor1', 'contingency_factor'),
                    ('length', 'length'),
                    ('r', 'r'),
                    ('angle_droop', 'angle_droop'),
                    ('min_firing_angle_f', 'min_firing_angle_f'),
                    ('max_firing_angle_f', 'max_firing_angle_f'),
                    ('min_firing_angle_t', 'min_firing_angle_t'),
                    ('max_firing_angle_t', 'max_firing_angle_t'),
                    ('overload_cost', 'overload_cost'), ]

            for entry in devices["HVDC Line"]:

                elm = HvdcLine()
                elm.bus_from = bus_dict[entry['bus_from']]
                elm.bus_to = bus_dict[entry['bus_to']]
                set_object_properties(elm, prop, entry)

                if "control_mode" in entry.keys():
                    elm.control_mode = hvdc_ctrl_dict[entry["control_mode"]]

                if has_profiles:
                    profile_entry = device_profiles_dict[elm.idtag]
                    elm.active_prof = decompress_array(profile_entry['active'])
                    elm.rate_prof = decompress_array(profile_entry['rate'])
                    elm.Pset_prof = decompress_array(profile_entry['Pset'])
                    elm.Vset_f_prof = decompress_array(profile_entry['vset_from'])
                    elm.Vset_t_prof = decompress_array(profile_entry['vset_to'])
                    elm.overload_cost_prof = decompress_array(profile_entry['overload_cost'])

                circuit.add_hvdc(elm)

        # fill x, y
        logger += circuit.fill_xy_from_lat_lon()
        return circuit
    else:
        logger.add('The Json structure does not have a Circuit inside the devices!')
        return MultiCircuit()


def parse_json_data_v2(data: dict, logger: Logger):
    """
    New Json parser
    :param data:
    :param logger:
    :return:
    """
    devices = data['devices']
    profiles = data['profiles']

    if DeviceType.CircuitDevice.value in devices.keys():

        dta = devices[DeviceType.CircuitDevice.value]
        circuit = MultiCircuit(name=str(dta['name']),
                               Sbase=float(dta['sbase']),
                               fbase=float(dta['fbase']),
                               idtag=str(dta['id']))

        jcircuit = devices["Circuit"]
        circuit.Sbase = jcircuit["sbase"]

        # Countries
        country_dict = dict()
        if 'Country' in devices.keys():
            elms = devices["Country"]
            for jentry in elms:
                elm = Country(idtag=str(jentry['id']),
                              code=str(jentry['code']),
                              name=str(jentry['name']))
                circuit.countries.append(elm)
                country_dict[elm.idtag] = elm
        else:
            elm = Country(idtag=None, code='Default', name='Default')
            circuit.countries.append(elm)

        # Areas
        areas_dict = dict()
        if 'Area' in devices.keys():
            elms = devices["Area"]
            for jentry in elms:
                elm = Area(idtag=str(jentry['id']),
                           code=str(jentry['code']),
                           name=str(jentry['name']))
                circuit.areas.append(elm)
                areas_dict[elm.idtag] = elm
        else:
            elm = Area(idtag=None, code='Default', name='Default')
            circuit.areas.append(elm)

        # Zones
        zones_dict = dict()
        if 'Zone' in devices.keys():
            elms = devices["Zone"]
            for jentry in elms:
                elm = Zone(idtag=str(jentry['id']),
                           code=str(jentry['code']),
                           name=str(jentry['name']))
                circuit.zones.append(elm)
                zones_dict[elm.idtag] = elm
        else:
            elm = Zone(idtag=None, code='Default', name='Default')
            circuit.zones.append(elm)

        # Substations
        substations_dict = dict()
        if 'Substation' in devices.keys():
            elms = devices["Substation"]
            for jentry in elms:
                elm = Substation(idtag=str(jentry['id']),
                                 code=str(jentry['code']),
                                 name=str(jentry['name']))
                circuit.substations.append(elm)
                substations_dict[elm.idtag] = elm
        else:
            elm = Substation(idtag=None, code='Default', name='Default')
            circuit.substations.append(elm)

        # buses
        bus_dict = dict()
        if 'Bus' in devices.keys():
            buses = devices["Bus"]
            for jentry in buses:

                area_id = str(jentry['area']) if 'area' in jentry.keys() else ''
                zone_id = str(jentry['zone']) if 'zone' in jentry.keys() else ''
                substation_id = str(jentry['substation']) if 'substation' in jentry.keys() else ''
                country_id = str(jentry['country']) if 'country' in jentry.keys() else ''

                if area_id in areas_dict.keys():
                    area = areas_dict[area_id]
                else:
                    area = circuit.areas[0]

                if zone_id in zones_dict.keys():
                    zone = zones_dict[zone_id]
                else:
                    zone = circuit.zones[0]

                if substation_id in substations_dict.keys():
                    substation = substations_dict[substation_id]
                else:
                    substation = circuit.substations[0]

                if country_id in country_dict.keys():
                    country = country_dict[country_id]
                else:
                    country = circuit.countries[0]

                bus = Bus(name=str(jentry['name']),
                          idtag=str(jentry['id']),
                          vnom=float(jentry['vnom']),
                          vmin=float(jentry['vmin']),
                          vmax=float(jentry['vmax']),
                          r_fault=float(jentry['rf']),
                          x_fault=float(jentry['xf']),
                          xpos=float(jentry['x']),
                          ypos=float(jentry['y']),
                          height=float(jentry['h']),
                          width=float(jentry['w']),
                          active=bool(jentry['active']),
                          is_slack=bool(jentry['is_slack']),
                          area=area,
                          zone=zone,
                          substation=substation,
                          country=country,
                          longitude=float(jentry['lon']),
                          latitude=float(jentry['lat']))

                bus_dict[jentry['id']] = bus
                circuit.add_bus(bus)

        if 'Generator' in devices.keys():
            generators = devices["Generator"]
            for jentry in generators:
                gen = Generator(name=str(jentry['name']),
                                idtag=str(jentry['id']),
                                active_power=float(jentry['p']),
                                power_factor=float(jentry['pf']),
                                voltage_module=float(jentry['vset']),
                                is_controlled=bool(jentry['is_controlled']),
                                Qmin=float(jentry['qmin']),
                                Qmax=float(jentry['qmax']),
                                Snom=float(jentry['snom']),
                                active=bool(jentry['active']),
                                p_min=float(jentry['pmin']),
                                p_max=float(jentry['pmax']),
                                op_cost=float(jentry['cost']),
                                )
                gen.bus = bus_dict[jentry['bus']]
                circuit.add_generator(gen.bus, gen)

        if 'Battery' in devices.keys():
            batteries = devices["Battery"]
            for jentry in batteries:
                gen = Battery(name=str(jentry['name']),
                              idtag=str(jentry['id']),
                              active_power=float(jentry['p']),
                              power_factor=float(jentry['pf']),
                              voltage_module=float(jentry['vset']),
                              is_controlled=bool(jentry['is_controlled']),
                              Qmin=float(jentry['qmin']),
                              Qmax=float(jentry['qmax']),
                              Snom=float(jentry['snom']),
                              active=bool(jentry['active']),
                              p_min=float(jentry['pmin']),
                              p_max=float(jentry['pmax']),
                              op_cost=float(jentry['cost']),
                              )
                gen.bus = bus_dict[jentry['bus']]
                circuit.add_battery(gen.bus, gen)

        if 'Load' in devices.keys():
            loads = devices["Load"]
            for jentry in loads:
                elm = Load(name=str(jentry['name']),
                           idtag=str(jentry['id']),
                           P=float(jentry['p']),
                           Q=float(jentry['q']),
                           active=bool(jentry['active']))
                elm.bus = bus_dict[jentry['bus']]
                circuit.add_load(elm.bus, elm)

        if "Shunt" in devices.keys():
            shunts = devices["Shunt"]
            for jentry in shunts:
                elm = Shunt(name=str(jentry['name']),
                            idtag=str(jentry['id']),
                            G=float(jentry['g']),
                            B=float(jentry['b']),
                            active=bool(jentry['active']))
                elm.bus = bus_dict[jentry['bus']]
                circuit.add_shunt(elm.bus, elm)

        if "Line" in devices.keys():
            lines = devices["Line"]
            for entry in lines:
                elm = Line(bus_from=bus_dict[entry['bus_from']],
                           bus_to=bus_dict[entry['bus_to']],
                           name=str(entry['name']),
                           idtag=str(entry['id']),
                           r=float(entry['r']),
                           x=float(entry['x']),
                           b=float(entry['b']),
                           rate=float(entry['rate']),
                           active=entry['active'],
                           length=float(entry['length']),
                           )
                circuit.add_line(elm)

        if "Transformer" in devices.keys() or "Transformer2w" in devices.keys():

            if "Transformer" in devices.keys():
                transformers = devices["Transformer"]
            elif "Transformer2w" in devices.keys():
                transformers = devices["Transformer2w"]
            else:
                raise Exception('Transformer key not found')

            for entry in transformers:
                elm = Transformer2W(bus_from=bus_dict[entry['bus_from']],
                                    bus_to=bus_dict[entry['bus_to']],
                                    name=str(entry['name']),
                                    idtag=str(entry['id']),
                                    r=float(entry['r']),
                                    x=float(entry['x']),
                                    g=float(entry['g']),
                                    b=float(entry['b']),
                                    rate=float(entry['rate']),
                                    active=bool(entry['active']),
                                    tap=float(entry['tap_module']),
                                    shift_angle=float(entry['tap_angle']),
                                    )
                circuit.add_transformer2w(elm)

        if "VSC" in devices.keys():
            vsc = devices["VSC"]

            # TODO: call correct_buses_connection()

        if "HVDC Line" in devices.keys():
            hvdc = devices["HVDC Line"]

        # fill x, y
        logger += circuit.fill_xy_from_lat_lon()

        return circuit

    else:
        logger.add('The Json structure does not have a Circuit inside the devices!')
        return MultiCircuit()


def parse_json(file_name) -> MultiCircuit:
    """
    Parse JSON file into Circuit
    :param file_name: 
    :return: GridCal MultiCircuit
    """
    data = json.load(open(file_name))

    return parse_json_data(data)


def save_json_file_v3(file_path, circuit: MultiCircuit, simulation_drivers=list()):
    """
    Save JSON file
    :param file_path: file path 
    :param circuit: GridCal MultiCircuit element
    :param simulation_drivers: List of Simulation Drivers
    """
    elements = dict()
    element_profiles = dict()
    units_dict = dict()
    logger = Logger()

    def add_to_dict(d, d2, key):
        if key in d.keys():
            d[key].append(d2)
        else:
            d[key] = [d2]

    def add_to_dict2(d, d2, key):
        if key not in d.keys():
            d[key] = d2

    # add the circuit
    elements[DeviceType.CircuitDevice.value] = circuit.get_properties_dict()
    units_dict[DeviceType.CircuitDevice.value] = circuit.get_units_dict()
    element_profiles[DeviceType.CircuitDevice.value] = circuit.get_profiles_dict()

    # add the areas
    for cls in [circuit.substations, circuit.zones, circuit.areas, circuit.countries]:
        for elm in cls:
            # pack the bus data into a dictionary
            add_to_dict(d=elements, d2=elm.get_properties_dict(), key=elm.device_type.value)
            add_to_dict(d=element_profiles, d2=convert_to_sparse(elm.get_profiles_dict()), key=elm.device_type.value)
            add_to_dict2(d=units_dict, d2=elm.get_units_dict(), key=elm.device_type.value)

    # add the buses
    for elm in circuit.buses:

        # pack the bus data into a dictionary
        add_to_dict(d=elements, d2=elm.get_properties_dict(), key=elm.device_type.value)
        add_to_dict(d=element_profiles, d2=convert_to_sparse(elm.get_profiles_dict()), key=elm.device_type.value)
        add_to_dict2(d=units_dict, d2=elm.get_units_dict(), key=elm.device_type.value)

        # pack all the elements within the bus
        devices = elm.loads + elm.controlled_generators + elm.static_generators + elm.batteries + elm.shunts
        for device in devices:
            add_to_dict(d=elements, d2=device.get_properties_dict(), key=device.device_type.value)
            add_to_dict(d=element_profiles, d2=convert_to_sparse(device.get_profiles_dict()), key=device.device_type.value)
            add_to_dict2(d=units_dict, d2=device.get_units_dict(), key=device.device_type.value)

    # branches
    for branch_list in circuit.get_branch_lists():
        for elm in branch_list:
            # pack the branch data into a dictionary
            add_to_dict(d=elements, d2=elm.get_properties_dict(), key=elm.device_type.value)
            add_to_dict(d=element_profiles, d2=convert_to_sparse(elm.get_profiles_dict()), key=elm.device_type.value)
            add_to_dict2(d=units_dict, d2=elm.get_units_dict(), key=elm.device_type.value)

    # results
    results = dict()
    for driver in simulation_drivers:
        if driver is not None:

            if driver.name == 'Power Flow':

                bus_data = dict()
                for i, elm in enumerate(circuit.buses):
                    bus_data[elm.idtag] = {'vm': np.abs(driver.results.voltage[i]),
                                           'va': np.angle(driver.results.voltage[i])}

                branch_data = dict()
                for i, elm in enumerate(circuit.get_branches_wo_hvdc()):
                    branch_data[elm.idtag] = {'p': driver.results.Sf[i].real,
                                              'q': driver.results.Sf[i].imag,
                                              'losses': driver.results.losses[i].real}

                for i, elm in enumerate(circuit.hvdc_lines):
                    branch_data[elm.idtag] = {'p': driver.results.hvdc_Pf[i].real,
                                              'q': 0,
                                              'losses': driver.results.hvdc_losses[i].real}

                results["power_flow"] = {'bus': bus_data,
                                         'branch': branch_data}

            elif driver.name == 'Time Series':

                bus_data = dict()
                for i, elm in enumerate(circuit.buses):
                    bus_data[elm.idtag] = {'vm': np.abs(driver.results.voltage[:, i]).tolist(),
                                           'va': np.angle(driver.results.voltage[:, i]).tolist()}

                branch_data = dict()
                for i, elm in enumerate(circuit.get_branches_wo_hvdc()):
                    branch_data[elm.idtag] = {'p': driver.results.Sf[:, i].real.tolist(),
                                              'q': driver.results.Sf[:, i].imag.tolist(),
                                              'losses': driver.results.losses[:, i].real.tolist()}

                for i, elm in enumerate(circuit.hvdc_lines):
                    branch_data[elm.idtag] = {'p': driver.results.hvdc_Pf[:, i].real,
                                              'q': 0,
                                              'losses': driver.results.hvdc_losses[:, i].real}

                results["time_series"] = {'bus': bus_data,
                                          'branch': branch_data}

    data = {'version': '3',
            'review': '1',
            'software': 'GridCal',
            'units': units_dict,
            'devices': elements,
            'profiles': element_profiles,
            'results': results}

    data_str = json.dumps(data, indent=True)

    # Save json to a text file
    text_file = open(file_path, "w")
    text_file.write(data_str)
    text_file.close()

    return logger


def save_json_file_v4(file_path, circuit: MultiCircuit, simulation_drivers=list()):
    pass


if __name__ == '__main__':

    import GridCal.Engine as gce

    # fname = '/home/santi/Documentos/Git/GitHub/GridCal/Grids_and_profiles/grids/IEEE 39+HVDC line.gridcal'
    # _circuit = gce.FileOpen(fname).open()
    # save_json_file_v3('/home/santi/Escritorio/IEEE 39+HVDC line.ejson3', _circuit)

    data = json.load(open('/home/santi/Escritorio/IEEE 39+HVDC line.ejson3'))
    parse_json_data_v3(data, Logger())



# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from typing import List, Any, Dict, Tuple
import json
import numpy as np
from GridCal.Engine.basic_structures import Logger, CompressedJsonStruct
import GridCal.Engine.Devices as dev
from GridCal.Engine.Core.multi_circuit import MultiCircuit
from GridCal.Engine.IO.raw_parser import get_psse_transformer_impedances


# ----------------------------------------------------------------------------------------------------------------------
def parse_circuit(circuit: MultiCircuit, block: CompressedJsonStruct):
    """

    :param circuit:
    :param block:
    :return:
    """
    # ["ic", "sbase", "rev", "xfrrat", "nxfrat", "basfrq", "title1", "title2"]
    data = block.get_dict_at(0)
    circuit.fBase = data['basfrq']
    circuit.Sbase = data['sbase']
    circuit.name = data['title1']
    circuit.comments = str(data['title1']) + "\n" + str(data['title2'])


def get_circuit_block(circuit: MultiCircuit, fields) -> CompressedJsonStruct:
    """

    :param circuit:
    :param fields:
    :return:
    """
    block = CompressedJsonStruct(fields=fields)
    block.declare_n_entries(1)
    block.set_at(0, 'basfrq', circuit.fBase)
    block.set_at(0, 'sbase', circuit.Sbase)
    block.set_at(0, 'title1', circuit.name)
    return block


# ----------------------------------------------------------------------------------------------------------------------
def parse_areas(circuit: MultiCircuit, block: CompressedJsonStruct) -> Dict[int, Any]:
    """
    Parse PSSe areas
    :param circuit: Multi-circuit
    :param block: block to parse
    :return: dictionary to match the PSSe area code to the actual area object
    """
    # "iarea", "isw", "pdes", "ptol", "arname"
    area_dict = dict()

    for i in range(block.get_row_number()):
        data = block.get_dict_at(i)
        area = dev.Area()
        area.code = data['iarea']
        # area.code = data['isw']  # swing bus -> 0
        # area.code = data['pdes']  # desired exchange (MW) -> 0
        # area.code = data['ptol']  # tolerance (MW) -> 99999
        area.name = data['arname']

        circuit.add_area(area)
        area_dict[area.code] = area

    return area_dict


def get_areas_block(circuit: MultiCircuit, fields) -> CompressedJsonStruct:
    """
    Generate PSSe areas from the GridCal Areas
    :param circuit: Multi-circuit
    :param fields:
    :return: CompressedJsonStruct block
    """
    block = CompressedJsonStruct(fields=fields)
    block.declare_n_entries(circuit.get_area_number())

    for i, area in enumerate(circuit.get_areas()):
        # "iarea", "isw", "pdes", "ptol", "arname"
        block.set_at(i, 'iarea', i + 1)
        block.set_at(i, 'isw', 0)
        block.set_at(i, 'pdes', 0)
        block.set_at(i, 'ptol', 99999)
        block.set_at(i, 'arname', area.name)

    return block


# ----------------------------------------------------------------------------------------------------------------------
def parse_zones(circuit: MultiCircuit, block: CompressedJsonStruct) -> Dict[int, Any]:
    """
    Parse PSSe areas
    :param circuit: Multi-circuit
    :param block: block to parse
    :return: dictionary to match the PSSe area code to the actual area object
    """
    # "izone", "zoname"
    area_zone = dict()

    for i in range(block.get_row_number()):
        data = block.get_dict_at(i)
        zone = dev.Zone()
        zone.code = data['izone']
        zone.name = data['zoname']

        circuit.add_zone(zone)
        area_zone[zone.code] = zone

    return area_zone


def get_zones_block(circuit: MultiCircuit, fields) -> CompressedJsonStruct:
    """
    Generate PSSe areas from the GridCal Areas
    :param circuit: Multi-circuit
    :param fields:
    :return: CompressedJsonStruct block
    """
    block = CompressedJsonStruct(fields=fields)
    block.declare_n_entries(circuit.get_zone_number())

    for i, area in enumerate(circuit.get_zones()):
        # "izone", "zoname"
        block.set_at(i, 'izone', i + 1)
        block.set_at(i, 'zoname', area.name)

    return block


# ----------------------------------------------------------------------------------------------------------------------
def parse_buses(circuit: MultiCircuit, block: CompressedJsonStruct) \
        -> Tuple[Dict[int, Any], Dict[Any, Tuple[int, int]]]:
    """

    :param circuit:
    :param block:
    :return:
    """
    # ["ibus", "name", "baskv", "ide", "area", "zone", "owner", "vm", "va", "nvhi", "nvlo", "evhi", "evlo"]
    bus_dict = dict()
    bus_area = dict()
    for i in range(block.get_row_number()):
        data = block.get_dict_at(i)
        bus = dev.Bus()
        bus.code = data['ibus']
        bus.name = data['name']
        bus.Vnom = data['baskv']

        if data['ide'] == 4:
            bus.active = False
        elif data['ide'] == 3:
            bus.is_slack = True

        # dictionary with the psse index and the bus object
        bus_dict[data['ibus']] = bus

        # keep the relation to the areas and zones to match properly later
        bus_area[bus] = (data['area'], data['zone'])

        # add to the circuit
        circuit.add_bus(bus)

    return bus_dict, bus_area


def get_buses_block(circuit: MultiCircuit, fields) -> Tuple[CompressedJsonStruct, Dict[Any, int]]:
    """

    :param circuit:
    :param fields:
    :return:
    """
    block = CompressedJsonStruct(fields=fields)
    block.declare_n_entries(circuit.get_bus_number())

    areas_dict = {elm: i+1 for i, elm in enumerate(circuit.get_areas())}
    zones_dict = {elm: i+1 for i, elm in enumerate(circuit.get_zones())}
    se_dict = {elm: i+1 for i, elm in enumerate(circuit.get_substations())}
    rev_bus_dict = dict()

    for i, bus in enumerate(circuit.get_buses()):
        # ["ibus", "name", "baskv", "ide", "area", "zone", "owner", "vm", "va", "nvhi", "nvlo", "evhi", "evlo"]

        bus_tpe = bus.determine_bus_type().value
        if not bus.active:
            bus_tpe = 4

        i_area = areas_dict[bus.area]
        i_zone = zones_dict[bus.zone]

        try:
            i_psse = int(bus.code)
        except:
            i_psse = i + 1

        block.set_at(i, 'ibus', i_psse)
        block.set_at(i, 'ibus', i_psse)
        block.set_at(i, 'name', bus.name[:12])
        block.set_at(i, 'baskv', bus.Vnom)
        block.set_at(i, 'ide', bus_tpe)
        block.set_at(i, 'area', i_area)
        block.set_at(i, 'zone', i_zone)
        block.set_at(i, 'owner', 1)

        block.set_at(i, 'vm', 1.0)
        block.set_at(i, 'va', 0)

        block.set_at(i, 'nvhi', 1.1)
        block.set_at(i, 'nvlo', 0.9)

        block.set_at(i, 'evhi', 1.1)
        block.set_at(i, 'evlo', 0.9)

        rev_bus_dict[bus] = i_psse

    return block, rev_bus_dict


# ----------------------------------------------------------------------------------------------------------------------
def get_loads_block(circuit: MultiCircuit, fields, rev_bus_dict: Dict[Any, int]) -> CompressedJsonStruct:
    """

    :param circuit:
    :param fields:
    :param rev_bus_dict: dictionary of buses and their assigned psse number
    :return:
    """
    block = CompressedJsonStruct(fields=fields)
    block.declare_n_entries(circuit.get_bus_number())

    areas_dict = {elm: i+1 for i, elm in enumerate(circuit.get_areas())}
    zones_dict = {elm: i+1 for i, elm in enumerate(circuit.get_zones())}

    i = 0
    for k, bus in enumerate(circuit.get_buses()):
        for k2, elm in enumerate(bus.loads):

            block.set_at(i, "ibus", rev_bus_dict[elm.bus])
            block.set_at(i,  "loadid", k2 + 1)
            block.set_at(i,  "stat", int(elm.active))
            block.set_at(i,  "area", areas_dict[elm.bus.area])
            block.set_at(i,  "zone", zones_dict[elm.bus.zone])
            block.set_at(i,  "pl", elm.P)
            block.set_at(i,  "ql", elm.Q)
            block.set_at(i,  "ip", elm.Ir)
            block.set_at(i,  "iq", elm.Ii)
            block.set_at(i,  "yp", elm.G)
            block.set_at(i,  "yq", elm.B)
            block.set_at(i,  "owner", 0)
            block.set_at(i,  "scale", 0)  # scale yes/no
            block.set_at(i,  "intrpt", 0)
            block.set_at(i,  "dgenp", 0)
            block.set_at(i,  "dgenq", 0)
            block.set_at(i,  "dgenm", 0)
            block.set_at(i,  "loadtype", 0)

            i += 1

    return block


def parse_loads(circuit: MultiCircuit, block: CompressedJsonStruct, buses_dict: Dict[int, Any]):
    """

    :param circuit:
    :param block:
    :param buses_dict:
    :return:
    """

    # ["ibus", "loadid", "stat", "area", "zone", "pl", "ql", "ip", "iq", "yp", "yq",
    # "owner", "scale", "intrpt", "dgenp", "dgenq", "dgenm", "loadtype"]

    for i in range(block.get_row_number()):
        data = block.get_dict_at(i)

        elm = dev.Load()

        elm.bus = buses_dict[data['ibus']]
        elm.code = "{0}_{1}".format(elm.bus.code, data['loadid'])
        elm.active = bool(data['stat'])

        elm.P = data['pl']
        elm.Q = data['ql']

        elm.Ir = data['ip']
        elm.Ii = data['iq']

        elm.G = data['yp']
        elm.B = data['yq']

        circuit.add_load(elm.bus, elm)


# ----------------------------------------------------------------------------------------------------------------------
def get_fixed_shunts_block(circuit: MultiCircuit, fields, rev_bus_dict: Dict[Any, int]) -> CompressedJsonStruct:
    """

    :param circuit:
    :param fields:
    :param rev_bus_dict: dictionary of buses and their assigned psse number
    :return:
    """
    block = CompressedJsonStruct(fields=fields)

    n = 0
    for k, bus in enumerate(circuit.get_buses()):
        for k2, elm in enumerate(bus.shunts):
            if not elm.is_controlled:
                n += 1

    block.declare_n_entries(n)

    # "ibus", "shntid", "stat", "gl", "bl"
    i = 0
    for k, bus in enumerate(circuit.get_buses()):
        for k2, elm in enumerate(bus.shunts):

            if not elm.is_controlled:
                block.set_at(i, "ibus", rev_bus_dict[elm.bus])
                block.set_at(i,  "shntid", k2 + 1)
                block.set_at(i,  "stat", int(elm.active))
                block.set_at(i,  "gl", elm.G)
                block.set_at(i,  "bl", elm.B)

                i += 1

    return block


def parse_fixed_shunts(circuit: MultiCircuit, block: CompressedJsonStruct, buses_dict: Dict[int, Any]):
    """

    :param circuit:
    :param block:
    :param buses_dict:
    :return:
    """

    # "ibus", "shntid", "stat", "gl", "bl"

    for i in range(block.get_row_number()):
        data = block.get_dict_at(i)

        elm = dev.Shunt()

        elm.bus = buses_dict[data['ibus']]
        elm.code = "{0}_{1}".format(elm.bus.code, data['shntid'])
        elm.active = bool(data['stat'])
        elm.is_controlled = False
        elm.G = data['gl']
        elm.B = data['bl']

        circuit.add_shunt(elm.bus, elm)


def parse_switched_shunts(circuit: MultiCircuit, block: CompressedJsonStruct, buses_dict: Dict[int, Any]):
    """

    :param circuit:
    :param block:
    :param buses_dict:
    :return:
    """

    # "ibus", "shntid", "modsw", "adjm", "stat", "vswhi", "vswlo", "swreg", "nreg",
    # "rmpct", "rmidnt", "binit", "s1", "n1", "b1", "s2", "n2", "b2", "s3", "n3", "b3",
    # "s4", "n4", "b4", "s5", "n5", "b5", "s6", "n6", "b6", "s7", "n7", "b7", "s8", "n8", "b8"

    for i in range(block.get_row_number()):
        data = block.get_dict_at(i)

        elm = dev.Shunt()

        elm.bus = buses_dict[data['ibus']]
        elm.code = "{0}_{1}".format(elm.bus.code, data['shntid'])
        elm.active = bool(data['stat'])
        elm.is_controlled = True

        g = 0.0
        if data['modsw'] in [1, 2]:
            b = data['binit'] * data['rmpct'] / 100.0
        else:
            b = data['binit']

        elm.G = g
        elm.B = b

        circuit.add_shunt(elm.bus, elm)


# ----------------------------------------------------------------------------------------------------------------------
def get_generators_block(circuit: MultiCircuit, fields, rev_bus_dict: Dict[Any, int]) -> CompressedJsonStruct:
    """

    :param circuit:
    :param fields:
    :param rev_bus_dict:
    :return:
    """
    block = CompressedJsonStruct(fields=fields)
    block.declare_n_entries(circuit.get_bus_number())

    for i, elm in enumerate(circuit.get_generators()):

        if "_" in elm.code:
            machine_id = elm.code.split("_")[1]
        else:
            machine_id = ""

        block.set_at(i, "ibus", rev_bus_dict[elm.bus])
        block.set_at(i,  "machid", machine_id)
        block.set_at(i,  "pg", elm.P)
        block.set_at(i,  "qg", 0)
        block.set_at(i,  "qt", elm.Qmax)
        block.set_at(i,  "qb", elm.Qmin)
        block.set_at(i,  "vs", elm.Vset)
        block.set_at(i,  "ireg", 0)
        block.set_at(i,  "nreg", 0)
        block.set_at(i,  "mbase", elm.Snom)
        block.set_at(i,  "zr", 0)
        block.set_at(i,  "zx", 0)
        block.set_at(i,  "rt", 0)
        block.set_at(i,  "xt", 0)
        block.set_at(i,  "gtap", 0)
        block.set_at(i,  "stat", int(elm.active))
        block.set_at(i,  "rmpct", 0)
        block.set_at(i,  "pt", elm.Pmax)
        block.set_at(i,  "pb", elm.Pmin)
        block.set_at(i,  "baslod", 0)
        block.set_at(i,  "o1", None)
        block.set_at(i,  "f1", None)
        block.set_at(i,  "o2", None)
        block.set_at(i,  "f2", None)
        block.set_at(i,  "o3", None)
        block.set_at(i,  "f3", None)
        block.set_at(i,  "o4", None)
        block.set_at(i,  "f4", None)
        block.set_at(i,  "wmod", 0)
        block.set_at(i,  "wpf", 0)

    return block


def parse_generators(circuit: MultiCircuit, block: CompressedJsonStruct, buses_dict: Dict[int, Any]):
    """

    :param circuit:
    :param block:
    :param buses_dict:
    :return:
    """
    # "ibus", "machid", "pg", "qg", "qt", "qb", "vs", "ireg", "nreg", "mbase",
    # "zr", "zx", "rt", "xt", "gtap", "stat", "rmpct", "pt", "pb", "baslod",
    # "o1", "f1", "o2", "f2", "o3", "f3", "o4", "f4", "wmod", "wpf"

    for i in range(block.get_row_number()):
        data = block.get_dict_at(i)

        elm = dev.Generator()

        elm.bus = buses_dict[data['ibus']]
        elm.active = bool(data['stat'])
        elm.code = "{0}_{1}".format(elm.bus.code, data['machid'])
        elm.P = data['pg']
        elm.Vset = data['vs']
        elm.pf = 0.8  # data['ql']

        elm.Qmin = data['qb']
        elm.Qmax = data['qt']
        elm.Snom = data['mbase']
        elm.Pmin = data['pb']
        elm.Pmax = data['pt']

        circuit.add_generator(elm.bus, elm)


# ----------------------------------------------------------------------------------------------------------------------
def get_ac_lines_block(circuit: MultiCircuit, fields, rev_bus_dict: Dict[Any, int]) -> CompressedJsonStruct:
    """

    :param circuit:
    :param fields:
    :param rev_bus_dict:
    :return:
    """
    block = CompressedJsonStruct(fields=fields)
    block.declare_n_entries(circuit.get_bus_number())

    for i, elm in enumerate(circuit.get_lines()):

        if "_" in elm.code:
            ckt = elm.code.split("_")[2]
        else:
            ckt = "1"

        block.set_at(i,  "ibus", rev_bus_dict[elm.bus_from])
        block.set_at(i,   "jbus", rev_bus_dict[elm.bus_to])
        block.set_at(i,   "ckt", ckt)
        block.set_at(i,   "rpu", elm.R)
        block.set_at(i,   "xpu", elm.X)
        block.set_at(i,   "bpu", elm.B)
        block.set_at(i,   "name", elm.name)
        block.set_at(i,   "rate1", elm.rate)
        block.set_at(i,   "rate2", elm.rate * elm.contingency_factor)
        block.set_at(i,   "rate3", elm.rate)
        block.set_at(i,   "rate4", elm.rate)
        block.set_at(i,   "rate5", elm.rate)
        block.set_at(i,   "rate6", elm.rate)
        block.set_at(i,   "rate7", elm.rate)
        block.set_at(i,   "rate8", elm.rate)
        block.set_at(i,   "rate9", elm.rate)
        block.set_at(i,   "rate10", elm.rate)
        block.set_at(i,   "rate11", elm.rate)
        block.set_at(i,   "rate12", elm.rate)
        block.set_at(i,   "gi", 0)
        block.set_at(i,   "bi", 0)
        block.set_at(i,   "gj", 0)
        block.set_at(i,   "bj", 0)
        block.set_at(i,   "stat", int(elm.active))
        block.set_at(i,   "met", None)
        block.set_at(i,   "len", elm.length)
        block.set_at(i,   "o1", None)
        block.set_at(i,   "f1", None)
        block.set_at(i,   "o2", None)
        block.set_at(i,   "f2", None)
        block.set_at(i,   "o3", None)
        block.set_at(i,   "f3", None)
        block.set_at(i,   "o4", None)
        block.set_at(i,   "f4", None)

        return block


def parse_ac_lines(circuit: MultiCircuit, block: CompressedJsonStruct, buses_dict: Dict[int, Any]):
    """

    :param circuit:
    :param block:
    :param buses_dict:
    :return:
    """
    # "ibus", "jbus", "ckt", "rpu", "xpu", "bpu", "name",
    # "rate1", "rate2", "rate3", "rate4", "rate5", "rate6", "rate7", "rate8", "rate9", "rate10", "rate11", "rate12",
    # "gi", "bi", "gj", "bj", "stat", "met", "len",
    # "o1", "f1", "o2", "f2", "o3", "f3", "o4", "f4"]

    for i in range(block.get_row_number()):
        data = block.get_dict_at(i)

        elm = dev.Line()

        elm.bus_from = buses_dict[data['ibus']]
        elm.bus_to = buses_dict[data['jbus']]
        elm.active = bool(data['stat'])

        code = "{0}_{1}_{2}".format(elm.bus_from.code, elm.bus_to.code, data['ckt'])
        elm.code = code

        elm.name = str(data['name']).strip()
        if elm.name == '':
            elm.name = code

        elm.length = data['len']

        elm.rate = data['rate1']
        if float(data['rate2']) != 0:
            elm.contingency_factor = float(data['rate1']) / float(data['rate2'])

        elm.R = data['rpu']
        elm.X = data['xpu']
        elm.B = data['bpu']

        circuit.add_line(elm)

        # add the lie compensations as shunt devices

        if data['gi'] != 0 or data['bi'] != 0:
            sh1 = dev.Shunt(name='Line compensation', code=code, G=data['gi'], B=data['bi'])
            circuit.add_shunt(elm.bus_from, sh1)

        if data['gj'] != 0 or data['bj'] != 0:
            sh1 = dev.Shunt(name='Line compensation', code=code, G=data['gj'], B=data['bj'])
            circuit.add_shunt(elm.bus_to, sh1)


# ----------------------------------------------------------------------------------------------------------------------
def get_twotermdc_block(circuit: MultiCircuit, fields, rev_bus_dict: Dict[Any, int]) -> CompressedJsonStruct:
    """

    :param circuit:
    :param fields:
    :param rev_bus_dict:
    :return:
    """
    block = CompressedJsonStruct(fields=fields)
    block.declare_n_entries(circuit.get_hvdc_number())

    for i, elm in enumerate(circuit.get_hvdc()):

        block.set_at(i, "name", elm.name)
        block.set_at(i,  "mdc", 1)
        block.set_at(i,  "rdc", elm.r)
        block.set_at(i,  "setvl", elm.Pset)
        block.set_at(i,  "vschd", None)
        block.set_at(i,  "vcmod", None)
        block.set_at(i,  "rcomp", None)
        block.set_at(i,  "delti", None)
        block.set_at(i,  "met", None)
        block.set_at(i,  "dcvmin", None)
        block.set_at(i,  "cccitmx", None)
        block.set_at(i,  "cccacc", None)
        block.set_at(i,  "ipr", rev_bus_dict[elm.bus_from])
        block.set_at(i,  "nbr", None)
        block.set_at(i,  "anmxr", elm.max_firing_angle_f)
        block.set_at(i,  "anmnr", elm.min_firing_angle_f)
        block.set_at(i,  "rcr", None)
        block.set_at(i,  "xcr", None)
        block.set_at(i,  "ebasr", None)
        block.set_at(i,  "trr", None)
        block.set_at(i,  "tapr", None)
        block.set_at(i,  "tmxr", None)
        block.set_at(i,  "tmnr", None)
        block.set_at(i,  "stpr", None)
        block.set_at(i,  "icr", None)
        block.set_at(i,  "ndr", None)
        block.set_at(i,  "ifr", None)
        block.set_at(i,  "itr", None)
        block.set_at(i,  "idr", None)
        block.set_at(i,  "xcapr", None)
        block.set_at(i,  "ipi", rev_bus_dict[elm.bus_to])
        block.set_at(i,  "nbi", None)
        block.set_at(i,  "anmxi", elm.max_firing_angle_t)
        block.set_at(i,  "anmni", elm.min_firing_angle_t)
        block.set_at(i,  "rci", None)
        block.set_at(i,  "xci", None)
        block.set_at(i,  "ebasi", None)
        block.set_at(i,  "tri", None)
        block.set_at(i,  "tapi", None)
        block.set_at(i,  "tmxi", None)
        block.set_at(i,  "tmni", None)
        block.set_at(i,  "stpi", None)
        block.set_at(i,  "ici", None)
        block.set_at(i,  "ndi", None)
        block.set_at(i,  "ifi", None)
        block.set_at(i,  "iti", None)
        block.set_at(i,  "idi", None)
        block.set_at(i,  "xcapi", None)

        return block


def parse_twotermdc_lines(circuit: MultiCircuit, block: CompressedJsonStruct, buses_dict: Dict[int, Any]):
    """

    :param circuit:
    :param block:
    :param buses_dict:
    :return:
    """
    # "name", "mdc", "rdc", "setvl", "vschd", "vcmod", "rcomp", "delti", "met",
    # "dcvmin", "cccitmx", "cccacc", "ipr", "nbr", "anmxr", "anmnr", "rcr", "xcr",
    # "ebasr", "trr", "tapr", "tmxr", "tmnr", "stpr", "icr", "ndr", "ifr", "itr", "idr",
    # "xcapr", "ipi", "nbi", "anmxi", "anmni", "rci", "xci", "ebasi", "tri", "tapi", "tmxi",
    # "tmni", "stpi", "ici", "ndi", "ifi", "iti", "idi", "xcapi"

    for i in range(block.get_row_number()):
        data = block.get_dict_at(i)

        bus_from = buses_dict[data['ipr']]
        bus_to = buses_dict[data['ipi']]
        code = "{0}_{1}_{2}".format(bus_from.code, bus_to.code, '1')
        name1 = str(data['name']).replace("'", "").replace('"', "").replace('/', '').strip()
        if name1 == '':
            name1 = code

        if data['mdc'] == 1 or data['mdc'] == 0:
            # SETVL is in MW
            specified_power = data['setvl']
        elif data['mdc'] == 2:
            # SETVL is in A, specified_power in MW
            specified_power = data['setvl'] * data['vschd'] / 1000.0
        else:
            # doesn't say, so zero
            specified_power = 0.0

        z_base = data['vschd'] * data['vschd'] / circuit.Sbase
        r_pu = data['rdc'] / z_base

        Vset_f = 1.0
        Vset_t = 1.0

        # set the HVDC line active
        active = bus_from.active and bus_to.active

        obj = dev.HvdcLine(bus_from=bus_from,  # Rectifier as of PSSe
                           bus_to=bus_to,  # inverter as of PSSe
                           active=active,
                           name=name1,
                           idtag=code,
                           Pset=specified_power,
                           Vset_f=Vset_f,
                           Vset_t=Vset_t,
                           rate=specified_power,
                           r=r_pu,
                           min_firing_angle_f=np.deg2rad(data['anmnr']),
                           max_firing_angle_f=np.deg2rad(data['anmxr']),
                           min_firing_angle_t=np.deg2rad(data['anmni']),
                           max_firing_angle_t=np.deg2rad(data['anmxi']))

        circuit.add_hvdc(obj)


# ----------------------------------------------------------------------------------------------------------------------
def get_transformers_block(circuit: MultiCircuit, fields, rev_bus_dict: Dict[Any, int]) -> CompressedJsonStruct:
    """

    :param circuit:
    :param fields:
    :param rev_bus_dict:
    :return:
    """
    block = CompressedJsonStruct(fields=fields)
    block.declare_n_entries(circuit.get_bus_number())

    for i, elm in enumerate(circuit.get_transformers2w()):

        if "_" in elm.code:
            ckt = elm.code.split("_")[2]
        else:
            ckt = "1"

        v1, v2 = elm.get_from_to_nominal_voltages()

        block.set_at(i,  "ibus", rev_bus_dict[elm.bus_from])
        block.set_at(i,   "jbus", rev_bus_dict[elm.bus_to])
        block.set_at(i,   "kbus", 0)
        block.set_at(i,   "ckt", str(ckt))
        block.set_at(i,   "cw", 1)
        block.set_at(i,   "cz", 1)
        block.set_at(i,   "cm", 1)
        block.set_at(i,   "mag1", 0)
        block.set_at(i,   "mag2", 0)
        block.set_at(i,   "nmet", 2)
        block.set_at(i,   "name", elm.name)
        block.set_at(i,   "stat", int(elm.active))
        block.set_at(i,   "o1", None)
        block.set_at(i,   "f1", None)
        block.set_at(i,   "o2", None)
        block.set_at(i,   "f2", None)
        block.set_at(i,   "o3", None)
        block.set_at(i,   "f3", None)
        block.set_at(i,   "o4", None)
        block.set_at(i,   "f4", None)
        block.set_at(i,   "vecgrp", "")
        block.set_at(i,   "zcod", None)
        block.set_at(i,   "r1_2", elm.R)
        block.set_at(i,   "x1_2", elm.X)
        block.set_at(i,   "sbase1_2", 100)
        block.set_at(i,   "r2_3", 0)
        block.set_at(i,   "x2_3", 0)
        block.set_at(i,   "sbase2_3", 0)
        block.set_at(i,   "r3_1", 0)
        block.set_at(i,   "x3_1", 0)
        block.set_at(i,   "sbase3_1", 0)
        block.set_at(i,   "vmstar", 0)
        block.set_at(i,   "anstar", 0)
        block.set_at(i,   "windv1", elm.tap_module)
        block.set_at(i,   "nomv1", 1.0)
        block.set_at(i,   "ang1", elm.angle)
        block.set_at(i,   "wdg1rate1", elm.rate)
        block.set_at(i,   "wdg1rate2", elm.rate)
        block.set_at(i,   "wdg1rate3", elm.rate)
        block.set_at(i,   "wdg1rate4", elm.rate)
        block.set_at(i,   "wdg1rate5", elm.rate)
        block.set_at(i,   "wdg1rate6", elm.rate)
        block.set_at(i,   "wdg1rate7", elm.rate)
        block.set_at(i,   "wdg1rate8", elm.rate)
        block.set_at(i,   "wdg1rate9", elm.rate)
        block.set_at(i,   "wdg1rate10", elm.rate)
        block.set_at(i,   "wdg1rate11", elm.rate)
        block.set_at(i,   "wdg1rate12", elm.rate)
        block.set_at(i,   "cod1", 0)
        block.set_at(i,   "cont1", 0)
        block.set_at(i,   "node1", 0)
        block.set_at(i,   "rma1", elm.tap_module)
        block.set_at(i,   "rmi1", elm.tap_module)
        block.set_at(i,   "vma1", 1.1)
        block.set_at(i,   "vmi1", 0.9)
        block.set_at(i,   "ntp1", None)
        block.set_at(i,   "tab1", 0)
        block.set_at(i,   "cr1", 0)
        block.set_at(i,   "cx1", 0)
        block.set_at(i,   "cnxa1", 0)
        block.set_at(i,   "windv2", 1.0)
        block.set_at(i,   "nomv2", v2)
        block.set_at(i,   "ang2", None)
        block.set_at(i,   "wdg2rate1", None)
        block.set_at(i,   "wdg2rate2", None)
        block.set_at(i,   "wdg2rate3", None)
        block.set_at(i,   "wdg2rate4", None)
        block.set_at(i,   "wdg2rate5", None)
        block.set_at(i,   "wdg2rate6", None)
        block.set_at(i,   "wdg2rate7", None)
        block.set_at(i,   "wdg2rate8", None)
        block.set_at(i,   "wdg2rate9", None)
        block.set_at(i,   "wdg2rate10", None)
        block.set_at(i,   "wdg2rate11", None)
        block.set_at(i,   "wdg2rate12", None)
        block.set_at(i,   "cod2", None)
        block.set_at(i,   "cont2", None)
        block.set_at(i,   "node2", None)
        block.set_at(i,   "rma2", None)
        block.set_at(i,   "rmi2", None)
        block.set_at(i,   "vma2", None)
        block.set_at(i,   "vmi2", None)
        block.set_at(i,   "ntp2", None)
        block.set_at(i,   "tab2", None)
        block.set_at(i,   "cr2", None)
        block.set_at(i,   "cx2", None)
        block.set_at(i,   "cnxa2", None)
        block.set_at(i,   "windv3", None)
        block.set_at(i,   "nomv3", None)
        block.set_at(i,   "ang3", None)
        block.set_at(i,   "wdg3rate1", None)
        block.set_at(i,   "wdg3rate2", None)
        block.set_at(i,   "wdg3rate3", None)
        block.set_at(i,   "wdg3rate4", None)
        block.set_at(i,   "wdg3rate5", None)
        block.set_at(i,   "wdg3rate6", None)
        block.set_at(i,   "wdg3rate7", None)
        block.set_at(i,   "wdg3rate8", None)
        block.set_at(i,   "wdg3rate9", None)
        block.set_at(i,   "wdg3rate10", None)
        block.set_at(i,   "wdg3rate11", None)
        block.set_at(i,   "wdg3rate12", None)
        block.set_at(i,   "cod3", None)
        block.set_at(i,   "cont3", None)
        block.set_at(i,   "node3", None)
        block.set_at(i,   "rma3", None)
        block.set_at(i,   "rmi3", None)
        block.set_at(i,   "vma3", None)
        block.set_at(i,   "vmi3", None)
        block.set_at(i,   "ntp3", None)
        block.set_at(i,   "tab3", None)
        block.set_at(i,   "cr3", None)
        block.set_at(i,   "cx3", None)
        block.set_at(i,   "cnxa3", None)

    return block


def parse_transformers(circuit: MultiCircuit, block: CompressedJsonStruct, buses_dict: Dict[int, Any], logger=Logger()):
    """

    :param circuit:
    :param block:
    :param buses_dict:
    :return:
    """
    # ibus	 "jbus"	 "kbus"	 "ckt"	 "cw"	 "cz"	 "cm"	 "mag1"	 "mag2"	 "nmet"
    # "name"	 "stat"	 "o1"	 "f1"	 "o2"	 "f2"	 "o3"	 "f3"	 "o4"	 "f4"
    # "vecgrp"	 "zcod"	 "r1_2"	 "x1_2"
    # "sbase1_2"	 "r2_3"	 "x2_3"	 "sbase2_3"
    # "r3_1"	 "x3_1"	 "sbase3_1"
    # "vmstar"	 "anstar"	 "windv1"
    # "nomv1"	 "ang1"
    # "wdg1rate1"	 "wdg1rate2"	 "wdg1rate3"	 "wdg1rate4"	 "wdg1rate5"	 "wdg1rate6"
    # "wdg1rate7"	 "wdg1rate8"	 "wdg1rate9"	 "wdg1rate10"	 "wdg1rate11"	 "wdg1rate12"
    # "cod1"	 "cont1"	 "node1"	 "rma1"	 "rmi1"	 "vma1"	 "vmi1"	 "ntp1"
    # "tab1"	 "cr1"	 "cx1"	 "cnxa1"	 "windv2"	 "nomv2"	 "ang2"
    # "wdg2rate1"	 "wdg2rate2"	 "wdg2rate3"	 "wdg2rate4"	 "wdg2rate5"	 "wdg2rate6"
    # "wdg2rate7"	 "wdg2rate8"	 "wdg2rate9"	 "wdg2rate10"	 "wdg2rate11"	 "wdg2rate12"
    # "cod2"	 "cont2"	 "node2"	 "rma2"	 "rmi2"	 "vma2"	 "vmi2"	 "ntp2"
    # "tab2"	 "cr2"	 "cx2"	 "cnxa2"	 "windv3"	 "nomv3"	 "ang3"
    # "wdg3rate1"	 "wdg3rate2"	 "wdg3rate3"	 "wdg3rate4"	 "wdg3rate5"	 "wdg3rate6"
    # "wdg3rate7"	 "wdg3rate8"	 "wdg3rate9"	 "wdg3rate10"	 "wdg3rate11"	 "wdg3rate12"
    # "cod3"	 "cont3"	 "node3"	 "rma3"	 "rmi3"	 "vma3"	 "vmi3"	 "ntp3"
    # "tab3"	 "cr3"	 "cx3"	 "cnxa3"

    for i in range(block.get_row_number()):
        data = block.get_dict_at(i)

        # get the buses
        bus_from = buses_dict[data['ibus']]
        bus_to = buses_dict[data['jbus']]

        code = "{0}_{1}_{2}".format(bus_from.code, bus_to.code, data['ckt'])

        if data['nomv1'] == 0:
            V1 = bus_from.Vnom
        else:
            V1 = data['nomv1']

        if data['nomv2'] == 0:
            V2 = bus_to.Vnom
        else:
            V2 = data['nomv2']

        r, x, g, b, tap_mod, tap_angle = get_psse_transformer_impedances(CW=data['cw'],
                                                                         CZ=data['cz'],
                                                                         CM=data['cm'],
                                                                         V1=V1,
                                                                         V2=V2,
                                                                         sbase=100,
                                                                         logger=logger,
                                                                         code=code,
                                                                         MAG1=data['mag1'],
                                                                         MAG2=data['mag2'],
                                                                         WINDV1=data['windv1'],
                                                                         WINDV2=data['windv2'],
                                                                         ANG1=data['ang1'],
                                                                         NOMV1=data['nomv1'],
                                                                         NOMV2=data['nomv2'],
                                                                         R1_2=data['r1_2'],
                                                                         X1_2=data['x1_2'],
                                                                         SBASE1_2=data['sbase1_2'])

        if float(data['wdg1rate2']) != 0:
            contingency_factor = float(data['wdg1rate1']) / float(data['wdg1rate2'])
        else:
            contingency_factor = 1.0

        elm = dev.Transformer2W(bus_from=bus_from,
                                bus_to=bus_to,
                                idtag=None,
                                code=code,
                                name=data['name'],
                                HV=V1,
                                LV=V2,
                                r=r,
                                x=x,
                                g=g,
                                b=b,
                                rate=data['wdg1rate1'],
                                contingency_factor=round(contingency_factor, 6),
                                tap=tap_mod,
                                shift_angle=tap_angle,
                                active=bool(data['stat']),
                                mttf=0,
                                mttr=0)

        circuit.add_transformer2w(elm)


# ----------------------------------------------------------------------------------------------------------------------


def get_rawx_structure():
    """
    This enumerates the expected structures inside the rawx JSON structure
    :return: dictionary {key: list of properties}
    """
    return {'caseid': ["ic", "sbase", "rev", "xfrrat", "nxfrat", "basfrq", "title1", "title2"],
            'general': ["thrshz", "pqbrak", "blowup", "maxisollvls", "camaxreptsln", "chkdupcntlbl"],
            'gauss': ["itmx", "accp", "accq", "accm", "tol"],
            'newton': ["itmxn", "accn", "toln", "vctolq", "vctolv", "dvlim", "ndvfct"],
            'adjust': ["adjthr", "acctap", "taplim", "swvbnd", "mxtpss", "mxswim"],
            'tysl': ["itmxty", "accty", "tolty"],
            'rating': ["irate", "name", "desc"],
            'bus': ["ibus", "name", "baskv", "ide", "area", "zone", "owner", "vm", "va", "nvhi", "nvlo", "evhi", "evlo"],
            'load': ["ibus", "loadid", "stat", "area", "zone", "pl", "ql", "ip", "iq", "yp", "yq", "owner", "scale", "intrpt", "dgenp", "dgenq", "dgenm", "loadtype"],
            'fixshunt': ["ibus", "shntid", "stat", "gl", "bl"],
            'generator': ["ibus", "machid", "pg", "qg", "qt", "qb", "vs", "ireg", "nreg", "mbase", "zr", "zx", "rt", "xt", "gtap", "stat", "rmpct", "pt", "pb", "baslod", "o1", "f1", "o2", "f2", "o3", "f3", "o4", "f4", "wmod", "wpf"],
            'acline': ["ibus", "jbus", "ckt", "rpu", "xpu", "bpu", "name", "rate1", "rate2", "rate3", "rate4", "rate5", "rate6", "rate7", "rate8", "rate9", "rate10", "rate11", "rate12", "gi", "bi", "gj", "bj", "stat", "met", "len", "o1", "f1", "o2", "f2", "o3", "f3", "o4", "f4"],
            'sysswd': ["ibus", "jbus", "ckt", "xpu", "rate1", "rate2", "rate3", "rate4", "rate5", "rate6", "rate7", "rate8", "rate9", "rate10", "rate11", "rate12", "stat", "nstat", "met", "stype", "name"],
            'transformer': ["ibus", "jbus", "kbus", "ckt", "cw", "cz", "cm", "mag1", "mag2", "nmet", "name", "stat", "o1", "f1", "o2", "f2", "o3", "f3", "o4", "f4", "vecgrp", "zcod", "r1_2", "x1_2", "sbase1_2", "r2_3", "x2_3", "sbase2_3", "r3_1", "x3_1", "sbase3_1", "vmstar", "anstar", "windv1", "nomv1", "ang1", "wdg1rate1", "wdg1rate2", "wdg1rate3", "wdg1rate4", "wdg1rate5", "wdg1rate6", "wdg1rate7", "wdg1rate8", "wdg1rate9", "wdg1rate10", "wdg1rate11", "wdg1rate12", "cod1", "cont1", "node1", "rma1", "rmi1", "vma1", "vmi1", "ntp1", "tab1", "cr1", "cx1", "cnxa1", "windv2", "nomv2", "ang2", "wdg2rate1", "wdg2rate2", "wdg2rate3", "wdg2rate4", "wdg2rate5", "wdg2rate6", "wdg2rate7", "wdg2rate8", "wdg2rate9", "wdg2rate10", "wdg2rate11", "wdg2rate12", "cod2", "cont2", "node2", "rma2", "rmi2", "vma2", "vmi2", "ntp2", "tab2", "cr2", "cx2", "cnxa2", "windv3", "nomv3", "ang3", "wdg3rate1", "wdg3rate2", "wdg3rate3", "wdg3rate4", "wdg3rate5", "wdg3rate6", "wdg3rate7", "wdg3rate8", "wdg3rate9", "wdg3rate10", "wdg3rate11", "wdg3rate12", "cod3", "cont3", "node3", "rma3", "rmi3", "vma3", "vmi3", "ntp3", "tab3", "cr3", "cx3", "cnxa3"],
            'area': ["iarea", "isw", "pdes", "ptol", "arname"],
            'twotermdc': ["name", "mdc", "rdc", "setvl", "vschd", "vcmod", "rcomp", "delti", "met", "dcvmin", "cccitmx", "cccacc", "ipr", "nbr", "anmxr", "anmnr", "rcr", "xcr", "ebasr", "trr", "tapr", "tmxr", "tmnr", "stpr", "icr", "ndr", "ifr", "itr", "idr", "xcapr", "ipi", "nbi", "anmxi", "anmni", "rci", "xci", "ebasi", "tri", "tapi", "tmxi", "tmni", "stpi", "ici", "ndi", "ifi", "iti", "idi", "xcapi"],
            'vscdc': ["name", "mdc", "rdc", "o1", "f1", "o2", "f2", "o3", "f3", "o4", "f4", "ibus1", "type1", "mode1", "dcset1", "acset1", "aloss1", "bloss1", "minloss1", "smax1", "imax1", "pwf1", "maxq1", "minq1", "vsreg1", "nreg1", "rmpct1", "ibus2", "type2", "mode2", "dcset2", "acset2", "aloss2", "bloss2", "minloss2", "smax2", "imax2", "pwf2", "maxq2", "minq2", "vsreg2", "nreg2", "rmpct2"],
            'impcor': ["itable", "tap", "refact", "imfact"],
            'ntermdc': ["name", "nconv", "ndcbs", "ndcln", "mdc", "vconv", "vcmod", "vconvn"],
            'ntermdcconv': ["name", "ib", "nbrdg", "angmx", "angmn", "rc", "xc", "ebas", "tr", "tap", "tpmx", "tpmn", "tstp", "setvl", "dcpf", "marg", "cnvcod"],
            'ntermdcbus': ["name", "idc", "ib", "area", "zone", "dcname", "idc2", "rgrnd", "owner"],
            'ntermdclink': ["name", "idc", "jdc", "dcckt", "met", "rdc", "ldc"],
            'msline': ["ibus", "jbus", "mslid", "met", "dum1", "dum2", "dum3", "dum4", "dum5", "dum6", "dum7", "dum8", "dum9"],
            'zone': ["izone", "zoname"],
            'iatrans': ["arfrom", "arto", "trid", "ptran"],
            'owner': ["iowner", "owname"],
            'facts': ["name", "ibus", "jbus", "mode", "pdes", "qdes", "vset", "shmx", "trmx", "vtmn", "vtmx", "vsmx", "imx", "linx", "rmpct", "owner", "set1", "set2", "vsref", "fcreg", "nreg", "mname"],
            'swshunt': ["ibus", "shntid", "modsw", "adjm", "stat", "vswhi", "vswlo", "swreg", "nreg", "rmpct", "rmidnt", "binit", "s1", "n1", "b1", "s2", "n2", "b2", "s3", "n3", "b3", "s4", "n4", "b4", "s5", "n5", "b5", "s6", "n6", "b6", "s7", "n7", "b7", "s8", "n8", "b8"],
            'gne': ["name", "model", "nterm", "bus1", "bus2", "nreal", "nintg", "nchar", "stat", "owner", "nmet", "real1", "real2", "real3", "real4", "real5", "real6", "real7", "real8", "real9", "real10", "intg1", "intg2", "intg3", "intg4", "intg5", "intg6", "intg7", "intg8", "intg9", "intg10", "char1", "char2", "char3", "char4", "char5", "char6", "char7", "char8", "char9", "char10"],
            'indmach': ["ibus", "imid", "stat", "sc", "dc", "area", "zone", "owner", "tc", "bc", "mbase", "ratekv", "pcode", "pset", "hconst", "aconst", "bconst", "dconst", "econst", "ra", "xa", "xm", "r1", "x1", "r2", "x2", "x3", "e1", "se1", "e2", "se2", "ia1", "ia2", "xamult"],
            'sub': ["isub", "name", "lati", "long", "srg"],
            'subnode': ["isub", "inode", "name", "ibus", "stat", "vm", "va"],
            'subswd': ["isub", "inode", "jnode", "swdid", "name", "type", "stat", "nstat", "xpu", "rate1", "rate2", "rate3"],
            'subterm': ["isub", "inode", "type", "eqid", "ibus", "jbus", "kbus"]}


def rawx_parse(file_name: str) -> [MultiCircuit, Logger]:
    """
    Parse a rawx file from PSSe
    :param file_name: file name
    :return: [MultiCircuit, Logger] instances
    """
    # read json file into dictionary
    data = json.load(open(file_name))

    # get structures
    struct = get_rawx_structure()
    circuit = MultiCircuit()
    logger = Logger()

    # get the data
    if 'network' in data.keys():
        data2 = data['network']
    else:
        logger.add_error('This is not a rawx json file :(')
        return circuit, logger

    # bus dictionary
    bus_dict = dict()
    areas_dict = dict()
    zones_dict = dict()
    bus_area = dict()
    for entry, fields in struct.items():
        if entry in data2.keys():

            # read the struct values
            block = CompressedJsonStruct(fields=data2[entry]['fields'],
                                         data=data2[entry]['data'])

            if entry == 'caseid':
                parse_circuit(circuit=circuit, block=block)

            elif entry == 'general':
                pass

            elif entry == 'gauss':
                pass

            elif entry == 'newton':
                pass

            elif entry == 'adjust':
                pass

            elif entry == 'tysl':
                pass

            elif entry == 'rating':
                pass

            elif entry == 'bus':
                bus_dict, bus_area = parse_buses(circuit=circuit, block=block)

            elif entry == 'load':
                parse_loads(circuit=circuit, block=block, buses_dict=bus_dict)

            elif entry == 'fixshunt':
                parse_fixed_shunts(circuit=circuit, block=block, buses_dict=bus_dict)

            elif entry == 'generator':
                parse_generators(circuit=circuit, block=block, buses_dict=bus_dict)

            elif entry == 'acline':
                parse_ac_lines(circuit=circuit, block=block, buses_dict=bus_dict)

            elif entry == 'sysswd':
                pass

            elif entry == 'transformer':
                parse_transformers(circuit=circuit, block=block, buses_dict=bus_dict)

            elif entry == 'area':
                areas_dict = parse_areas(circuit=circuit, block=block)

            elif entry == 'twotermdc':
                parse_twotermdc_lines(circuit=circuit, block=block, buses_dict=bus_dict)

            elif entry == 'vscdc':
                pass

            elif entry == 'impcor':
                pass

            elif entry == 'ntermdc':
                pass

            elif entry == 'ntermdcconv':
                pass

            elif entry == 'ntermdcbus':
                pass

            elif entry == 'ntermdclink':
                pass

            elif entry == 'msline':
                pass

            elif entry == 'zone':
                zones_dict = parse_zones(circuit=circuit, block=block)

            elif entry == 'iatrans':
                pass

            elif entry == 'owner':
                pass

            elif entry == 'facts':
                pass

            elif entry == 'swshunt':
                parse_switched_shunts(circuit=circuit, block=block, buses_dict=bus_dict)

            elif entry == 'gne':
                pass

            elif entry == 'indmach':
                pass

            elif entry == 'sub':
                pass

            elif entry == 'subnode':
                pass

            elif entry == 'subswd':
                pass

            elif entry == 'subterm':
                pass

            else:
                logger.add_warning('Unkown rawx structure ' + entry)

        else:
            logger.add_warning(entry + " not found")

    # force the buses parsing
    # this piece is here to ensure that all the elements are already present
    for bus, (id_area, id_zone) in bus_area.items():
        bus.area = areas_dict[id_area]
        bus.zone = zones_dict[id_zone]

    return circuit, logger


def rawx_writer(file_name: str, circuit: MultiCircuit) -> Logger:
    """
    RAWx export
    :param file_name: file name to save to
    :param circuit: MultiCircuit instance
    :return: Logger instance
    """
    struct = get_rawx_structure()
    logger = Logger()

    data = dict()
    rev_bus_dict = dict()
    for entry, fields in struct.items():

        # default structure
        block = CompressedJsonStruct(fields=fields)

        # fill the structure accordingly
        if entry == 'caseid':
            block = get_circuit_block(circuit=circuit, fields=fields)

        elif entry == 'general':
            # this is fixed
            block.set_data([0.0001, 0.7, 5.0, 4, 20, 0])

        elif entry == 'gauss':
            # this is fixed
            block.set_data([100, 1.6, 1.6, 1.0, 0.0001])

        elif entry == 'newton':
            # this is fixed
            block.set_data([100, 0.25, 0.01, 0.1, 0.00001, 0.99, 0.99])

        elif entry == 'adjust':
            # this is fixed
            block.set_data([0.005, 1.0, 0.05, 100.0, 99, 10])

        elif entry == 'tysl':
            # this is fixed
            block.set_data([20, 1.0, 0.00001])

        elif entry == 'rating':

            # this is fixed
            block.set_data([[1, "RATE1", "RATING SET 1"],
                                 [2, "RATE2", "RATING SET 2"],
                                 [3, "RATE3", "RATING SET 3"],
                                 [4, "RATE4", "RATING SET 4"],
                                 [5, "RATE5", "RATING SET 5"],
                                 [6, "RATE6", "RATING SET 6"],
                                 [7, "RATE7", "RATING SET 7"],
                                 [8, "RATE8", "RATING SET 8"],
                                 [9, "RATE9", "RATING SET 9"],
                                 [10, "RATE10", "RATING SET 10"],
                                 [11, "RATE11", "RATING SET 11"],
                                 [12, "RATE12", "RATING SET 12"]])

        elif entry == 'bus':
            block, rev_bus_dict = get_buses_block(circuit=circuit, fields=fields)

        elif entry == 'load':
            block = get_loads_block(circuit=circuit, fields=fields, rev_bus_dict=rev_bus_dict)

        elif entry == 'fixshunt':
            block = get_fixed_shunts_block(circuit=circuit, fields=fields, rev_bus_dict=rev_bus_dict)

        elif entry == 'generator':
            block = get_generators_block(circuit=circuit, fields=fields, rev_bus_dict=rev_bus_dict)

        elif entry == 'acline':
            block = get_ac_lines_block(circuit=circuit, fields=fields, rev_bus_dict=rev_bus_dict)

        elif entry == 'sysswd':
            pass

        elif entry == 'transformer':
            block = get_transformers_block(circuit=circuit, fields=fields, rev_bus_dict=rev_bus_dict)

        elif entry == 'area':
            block = get_areas_block(circuit=circuit, fields=fields)

        elif entry == 'twotermdc':
            block = get_twotermdc_block(circuit=circuit, fields=fields, rev_bus_dict=rev_bus_dict)

        elif entry == 'vscdc':
            pass

        elif entry == 'impcor':
            pass

        elif entry == 'ntermdc':
            pass

        elif entry == 'ntermdcconv':
            pass

        elif entry == 'ntermdcbus':
            pass

        elif entry == 'ntermdclink':
            pass

        elif entry == 'msline':
            pass

        elif entry == 'zone':
            block = get_zones_block(circuit=circuit, fields=fields)

        elif entry == 'iatrans':
            pass

        elif entry == 'owner':
            block.set_data([1, "Default"])

        elif entry == 'facts':
            pass

        elif entry == 'swshunt':
            pass

        elif entry == 'gne':
            pass

        elif entry == 'indmach':
            pass

        elif entry == 'sub':
            pass

        elif entry == 'subnode':
            pass

        elif entry == 'subswd':
            pass

        elif entry == 'subterm':
            pass

        else:
            logger.add_warning('Unkown rawx structure ' + entry)

        # get the dictionary
        data[entry] = block.get_final_dict()

    # save the data into a file
    rawx = {'network': data}
    with open(file_name, 'w') as fp:
        fp.write(json.dumps(rawx, indent=True))

    return logger



# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import pandas as pd
import numpy as np
from matplotlib import pyplot as plt

from GridCal.Engine.Devices.bus import Bus
from GridCal.Engine.Devices.enumerations import BranchType, ConverterControlType

from GridCal.Engine.Devices.editable_device import EditableDevice, DeviceType, GCProp


class UPFC(EditableDevice):

    def __init__(self, bus_from: Bus = None, bus_to: Bus = None, name='UPFC', code='', idtag=None, active=True,
                 rs=0.0, xs=0.00001, rl=0.0, xl=0.0, bl=0.0, rp=0.0, xp=0.0, vp=1.0, Pset = 0.0, Qset=0.0, rate=9999,
                 mttf=0, mttr=0, cost=100, cost_prof=None, rate_prof=None, active_prof=None, contingency_factor=1.0,
                 contingency_enabled=True, monitor_loading=True, contingency_factor_prof=None):
        """
        Unified Power Flow Converter (UPFC)
        :param bus_from:
        :param bus_to:
        :param name:
        :param idtag:
        :param active:
        :param rs: series resistance (p.u.)
        :param xs: series reactance (p.u.)
        :param rl: line resistance (p.u.)
        :param xl: line reactance (p.u.)
        :param bl: line shunt susceptance (p.u.)
        :param rp: shunt resistance (p.u.)
        :param xp: shunt reactance (p.u.)
        :param vp: shunt voltage set point (p.u.)
        :param rate: Power rating (MVA)
        :param Pset: Power set point (MW)
        :param mttf:
        :param mttr:
        :param cost:
        :param cost_prof:
        :param rate_prof:
        :param active_prof:
        """

        EditableDevice.__init__(self,
                                name=name,
                                idtag=idtag,
                                active=active,
                                code=code,
                                device_type=DeviceType.UpfcDevice,
                                editable_headers={'name': GCProp('', str, 'Name of the branch.'),
                                                  'idtag': GCProp('', str, 'Unique ID'),
                                                  'code': GCProp('', str, 'Secondary ID'),
                                                  'bus_from': GCProp('', DeviceType.BusDevice,
                                                                     'Name of the bus at the "from" side of the branch.'),
                                                  'bus_to': GCProp('', DeviceType.BusDevice,
                                                                   'Name of the bus at the "to" side of the branch.'),
                                                  'active': GCProp('', bool, 'Is the branch active?'),
                                                  'rate': GCProp('MVA', float, 'Thermal rating power of the branch.'),

                                                  'contingency_factor': GCProp('p.u.', float,
                                                                               'Rating multiplier for contingencies.'),
                                                  'contingency_enabled': GCProp('', bool,
                                                                                'Consider this UPFC for contingencies.'),
                                                  'monitor_loading': GCProp('', bool,
                                                                            'Monitor this device loading for optimization, NTC or contingency studies.'),
                                                  'mttf': GCProp('h', float, 'Mean time to failure, '
                                                                 'used in reliability studies.'),
                                                  'mttr': GCProp('h', float, 'Mean time to recovery, '
                                                                 'used in reliability studies.'),
                                                  'Rl': GCProp('p.u.', float, 'Line resistance.'),
                                                  'Xl': GCProp('p.u.', float, 'Line reactance.'),
                                                  'Bl': GCProp('p.u.', float, 'Line susceptance.'),
                                                  'Rs': GCProp('p.u.', float, 'Series resistance.'),
                                                  'Xs': GCProp('p.u.', float, 'Series reactance.'),
                                                  'Rsh': GCProp('p.u.', float, 'Shunt resistance.'),
                                                  'Xsh': GCProp('p.u.', float, 'Shunt resistance.'),
                                                  'Vsh': GCProp('p.u.', float, 'Shunt voltage set point.'),
                                                  'Pfset': GCProp('MW', float, 'Active power set point.'),
                                                  'Qfset': GCProp('MVAr', float, 'Active power set point.'),
                                                  'Cost': GCProp('e/MWh', float, 'Cost of overloads. Used in OPF.'),
                                                  },
                                non_editable_attributes=['bus_from', 'bus_to', 'idtag'],
                                properties_with_profile={'active': 'active_prof',
                                                         'rate': 'rate_prof',
                                                         'contingency_factor': 'contingency_factor_prof',
                                                         'Cost': 'Cost_prof'})

        self.bus_from = bus_from
        self.bus_to = bus_to

        # List of measurements
        self.measurements = list()

        # total impedance and admittance in p.u.
        self.Rl = rl
        self.Xl = xl
        self.Bl = bl
        self.Rs = rs
        self.Xs = xs
        self.Rsh = rp
        self.Xsh = xp
        self.Vsh = vp
        self.Pfset = Pset
        self.Qfset = Qset

        self.Cost = cost
        self.Cost_prof = cost_prof

        self.mttf = mttf
        self.mttr = mttr

        self.active = active
        self.active_prof = active_prof

        # branch rating in MVA
        self.rate = rate
        self.contingency_factor = contingency_factor
        self.contingency_enabled: bool = contingency_enabled
        self.monitor_loading: bool = monitor_loading
        self.rate_prof = rate_prof
        self.contingency_factor_prof = contingency_factor_prof

        # branch type: Line, Transformer, etc...
        self.branch_type = BranchType.UPFC

    def change_base(self, Sbase_old, Sbase_new):
        b = Sbase_new / Sbase_old

        self.Rl *= b
        self.Xl *= b
        self.Bl *= b
        self.Rs *= b
        self.Xs *= b
        self.Rsh *= b
        self.Xs *= b

    def get_properties_dict(self, version=3):
        """
        Get json dictionary
        :return:
        """
        if version == 2:
            return {'id': self.idtag,
                    'type': 'upfc',
                    'phases': 'ps',
                    'name': self.name,
                    'name_code': self.code,
                    'bus_from': self.bus_from.idtag,
                    'bus_to': self.bus_to.idtag,
                    'active': self.active,
                    'rate': self.rate,
                    'rl': self.Rl,
                    'xl': self.Xl,
                    'bl': self.Bl,
                    'rs': self.Rs,
                    'xs': self.Xs,
                    'rsh': self.Rsh,
                    'xsh': self.Xsh,
                    'vsh': self.Vsh,
                    'Pfset': self.Pfset,
                    'Qfset': self.Qfset
                    }
        elif version == 3:
            return {'id': self.idtag,
                    'type': 'upfc',
                    'phases': 'ps',
                    'name': self.name,
                    'name_code': self.code,
                    'bus_from': self.bus_from.idtag,
                    'bus_to': self.bus_to.idtag,
                    'active': self.active,
                    'rate': self.rate,
                    'contingency_factor1': self.contingency_factor,
                    'contingency_factor2': self.contingency_factor,
                    'contingency_factor3': self.contingency_factor,
                    'rl': self.Rl,
                    'xl': self.Xl,
                    'bl': self.Bl,
                    'rs': self.Rs,
                    'xs': self.Xs,
                    'rsh': self.Rsh,
                    'xsh': self.Xsh,
                    'vsh': self.Vsh,
                    'Pfset': self.Pfset,
                    'Qfset': self.Qfset
                    }
        else:
            return dict()

    def get_profiles_dict(self, version=3):
        """

        :return:
        """
        if self.active_prof is not None:
            active_prof = self.active_prof.tolist()
            rate_prof = self.rate_prof.tolist()
        else:
            active_prof = list()
            rate_prof = list()

        return {'id': self.idtag,
                'active': active_prof,
                'rate': rate_prof}

    def get_units_dict(self, version=3):
        """
        Get units of the values
        """
        return {'rate': 'MW',
                'r': 'p.u.',
                'x': 'p.u.',
                'b': 'p.u.',
                'g': 'p.u.'}

    def get_coordinates(self):
        """
        Get the branch defining coordinates
        """
        return [self.bus_from.get_coordinates(), self.bus_to.get_coordinates()]

    def plot_profiles(self, time_series=None, my_index=0, show_fig=True):
        """
        Plot the time series results of this object
        :param time_series: TimeSeries Instance
        :param my_index: index of this object in the simulation
        :param show_fig: Show the figure?
        """

        if time_series is not None:
            fig = plt.figure(figsize=(12, 8))

            ax_1 = fig.add_subplot(211)
            ax_2 = fig.add_subplot(212, sharex=ax_1)

            x = time_series.results.time

            # loading
            y = time_series.results.loading.real * 100.0
            df = pd.DataFrame(data=y[:, my_index], index=x, columns=[self.name])
            ax_1.set_title('Loading', fontsize=14)
            ax_1.set_ylabel('Loading [%]', fontsize=11)
            df.plot(ax=ax_1)

            # losses
            y = np.abs(time_series.results.losses)
            df = pd.DataFrame(data=y[:, my_index], index=x, columns=[self.name])
            ax_2.set_title('Losses', fontsize=14)
            ax_2.set_ylabel('Losses [MVA]', fontsize=11)
            df.plot(ax=ax_2)

            plt.legend()
            fig.suptitle(self.name, fontsize=20)

        if show_fig:
            plt.show()



# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
import numpy as np
import scipy.sparse as sp
import GridCal.Engine.Core.topology as tp


class UpfcData:

    def __init__(self, nelm, nbus, ntime=1):
        """

        :param nelm:
        :param nbus:
        :param ntime:
        """
        self.nelm = nelm
        self.ntime = ntime

        self.names = np.zeros(nelm, dtype=object)
        self.Rl = np.zeros(nelm)
        self.Xl = np.zeros(nelm)
        self.Bl = np.zeros(nelm)

        self.Rs = np.zeros(nelm)
        self.Xs = np.zeros(nelm)

        self.Rsh = np.zeros(nelm)
        self.Xsh = np.zeros(nelm)

        self.active = np.zeros((nelm, ntime), dtype=int)
        self.Vsh = np.zeros((nelm, ntime))
        self.Pset = np.zeros((nelm, ntime))
        self.Qset = np.zeros((nelm, ntime))

        self.C_elm_bus = sp.lil_matrix((nelm, nbus), dtype=int)  # this ons is just for splitting islands

    def slice(self, elm_idx, bus_idx, time_idx=None):
        """

        :param elm_idx:
        :param bus_idx:
        :param time_idx:
        :return:
        """

        if time_idx is None:
            tidx = elm_idx
        else:
            tidx = np.ix_(elm_idx, time_idx)

        data = UpfcData(nelm=len(elm_idx), nbus=len(bus_idx))

        data.names = self.names[elm_idx]
        data.Rl = self.Rl[elm_idx]
        data.Xl = self.Xl[elm_idx]
        data.Bl = self.Bl[elm_idx]

        data.Rs = self.Rs[elm_idx]
        data.Xs = self.Xs[elm_idx]

        data.Rsh = self.Rsh[elm_idx]
        data.Xsh = self.Xsh[elm_idx]

        data.active = self.active[tidx]
        data.Pset = self.Pset[tidx]
        data.Qset = self.Qset[tidx]
        data.Vsh = self.Vsh[tidx]

        data.C_elm_bus = self.C_elm_bus[np.ix_(elm_idx, bus_idx)]

        return data

    def get_island(self, bus_idx, t_idx=0):
        """
        Get the elements of the island given the bus indices
        :param bus_idx: list of bus indices
        :return: list of line indices of the island
        """
        if self.nelm:
            return tp.get_elements_of_the_island(self.C_elm_bus, bus_idx, active=self.active[:, t_idx])
        else:
            return np.zeros(0, dtype=int)

    def __len__(self):
        return self.nelm



# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
import io
from enum import Enum
from datetime import datetime
from PySide2 import QtCore, QtGui, QtWidgets

from GridCal.Engine.basic_structures import Logger
from GridCal.Gui.GuiFunctions import ObjectsModel, get_tree_model, get_list_model


class ProfileTypes(Enum):
    Loads = 1,
    Generators = 2


class NewProfilesStructureDialogue(QtWidgets.QDialog):
    """
    New profile dialogue window
    """
    def __init__(self):
        super(NewProfilesStructureDialogue, self).__init__()
        self.setObjectName("self")
        # self.resize(200, 71)
        # self.setMinimumSize(QtCore.QSize(200, 71))
        # self.setMaximumSize(QtCore.QSize(200, 71))
        self.setContextMenuPolicy(QtCore.Qt.NoContextMenu)
        # icon = QtGui.QIcon()
        # icon.addPixmap(QtGui.QPixmap("Icons/Plus-32.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        # self.setWindowIcon(icon)
        self.layout = QtWidgets.QVBoxLayout(self)

        # calendar
        self.calendar = QtWidgets.QDateTimeEdit()
        d = datetime.today()
        self.calendar.setDateTime(QtCore.QDateTime(d.year, 1, 1, 00, 00, 00))

        # number of time steps
        self.steps_spinner = QtWidgets.QSpinBox()
        self.steps_spinner.setMinimum(1)
        self.steps_spinner.setMaximum(9999999)
        self.steps_spinner.setValue(1)

        # time step length
        self.step_length = QtWidgets.QDoubleSpinBox()
        self.step_length.setMinimum(1)
        self.step_length.setMaximum(60)
        self.step_length.setValue(1)

        # units combo box
        self.units = QtWidgets.QComboBox()
        self.units.setModel(get_list_model(['h', 'm', 's']))

        # accept button
        self.accept_btn = QtWidgets.QPushButton()
        self.accept_btn.setText('Accept')
        self.accept_btn.clicked.connect(self.accept_click)

        # labels

        # add all to the GUI
        self.layout.addWidget(QtWidgets.QLabel("Start date"))
        self.layout.addWidget(self.calendar)

        self.layout.addWidget(QtWidgets.QLabel("Number of time steps"))
        self.layout.addWidget(self.steps_spinner)

        self.layout.addWidget(QtWidgets.QLabel("Time step length"))
        self.layout.addWidget(self.step_length)

        self.layout.addWidget(QtWidgets.QLabel("Time units"))
        self.layout.addWidget(self.units)

        self.layout.addWidget(self.accept_btn)

        self.setLayout(self.layout)

        self.setWindowTitle('New profiles structure')

    def accept_click(self):
        self.accept()

    def get_values(self):
        steps = self.steps_spinner.value()

        step_length = self.step_length.value()

        step_unit = self.units.currentText()

        time_base = self.calendar.dateTime()

        return steps, step_length, step_unit, time_base.toPython()


def fill_tree_from_logs(logger: Logger):
    """
    Fill logger tree
    :param logger: Logger instance
    :return: QStandardItemModel instance
    """
    d = logger.to_dict()
    editable = False
    model = QtGui.QStandardItemModel()
    model.setHorizontalHeaderLabels(['Time', 'Element', 'Value', 'Expected value'])
    parent = model.invisibleRootItem()

    for severity, messages_dict in d.items():
        severity_child = QtGui.QStandardItem(severity)

        # print(severity)

        for message, data_list in messages_dict.items():
            message_child = QtGui.QStandardItem(message)

            # print('\t', message)

            for time, elm, value, expected_value in data_list:

                # print('\t', '\t', time, elm, value, expected_value)

                time_child = QtGui.QStandardItem(time)
                time_child.setEditable(editable)

                elm_child = QtGui.QStandardItem(elm)
                elm_child.setEditable(editable)

                value_child = QtGui.QStandardItem(value)
                value_child.setEditable(editable)

                expected_val_child = QtGui.QStandardItem(expected_value)
                expected_val_child.setEditable(editable)

                message_child.appendRow([time_child, elm_child, value_child, expected_val_child])

            message_child.setEditable(editable)

            severity_child.appendRow(message_child)

        severity_child.setEditable(editable)
        parent.appendRow(severity_child)

    return model


class MTreeExpandHook(QtCore.QObject):
    """
    MTreeExpandHook( QTreeView )
    """

    def __init__(self, tree):
        super(MTreeExpandHook, self).__init__()
        self.setParent(tree)
        # NOTE viewport for click event listen
        tree.viewport().installEventFilter(self)
        self.tree = tree

    def eventFilter(self, receiver, event):
        if (
            # NOTE mouse left click
            event.type() == QtCore.QEvent.Type.MouseButtonPress
            # NOTE keyboard shift press
            and event.modifiers() & QtCore.Qt.ShiftModifier
        ):
            # NOTE get mouse local position
            pos = self.tree.mapFromGlobal(QtGui.QCursor.pos())
            index = self.tree.indexAt(pos)
            if not self.tree.isExpanded(index):
                # NOTE expand all child
                self.tree.expandRecursively(index)
                return True
        return super(MTreeExpandHook, self).eventFilter(self.tree, event)


class LogsDialogue(QtWidgets.QDialog):
    """
    New profile dialogue window
    """
    def __init__(self, name, logger: Logger(), expand_all=True):
        super(LogsDialogue, self).__init__()
        self.setObjectName("self")
        self.setContextMenuPolicy(QtCore.Qt.NoContextMenu)
        self.main_layout = QtWidgets.QVBoxLayout(self)

        self.logger = logger

        # logs_list
        self.logs_table = QtWidgets.QTreeView()
        model = fill_tree_from_logs(logger)
        self.logs_table.setModel(model)
        self.logs_table.setFirstColumnSpanned(0, QtCore.QModelIndex(), True)
        self.logs_table.setFirstColumnSpanned(1, QtCore.QModelIndex(), True)
        self.logs_table.setAnimated(True)
        # MTreeExpandHook(self.logs_table)

        if expand_all:
            self.logs_table.expandAll()

        # accept button
        self.accept_btn = QtWidgets.QPushButton()
        self.accept_btn.setText('Accept')
        self.accept_btn.clicked.connect(self.accept_click)

        self.save_btn = QtWidgets.QPushButton()
        self.save_btn.setText('Save')
        self.save_btn.clicked.connect(self.save_click)

        self.copy_btn = QtWidgets.QPushButton()
        self.copy_btn.setText('Copy')
        self.copy_btn.clicked.connect(self.copy_click)

        self.btn_frame = QtWidgets.QFrame()
        self.btn_layout = QtWidgets.QHBoxLayout(self.btn_frame)
        self.btn_spacer = QtWidgets.QSpacerItem(10, 10, QtWidgets.QSizePolicy.Expanding)
        self.btn_layout.addWidget(self.save_btn)
        self.btn_layout.addWidget(self.copy_btn)
        self.btn_layout.addSpacerItem(self.btn_spacer)
        self.btn_layout.addWidget(self.accept_btn)
        self.btn_frame.setLayout(self.btn_layout)

        # add all to the GUI
        self.main_layout.addWidget(self.logs_table)
        self.main_layout.addWidget(self.btn_frame)

        self.setLayout(self.main_layout)

        self.setWindowTitle(name)

        h = 400
        self.resize(int(1.61 * h), h)

    def accept_click(self):
        """
        Accept and close
        """
        self.accept()

    def save_click(self):
        """
        Save the logs to excel or CSV
        """
        file, filter_ = QtWidgets.QFileDialog.getSaveFileName(self, "Export results", '',
                                                              filter="CSV (*.csv);;Excel files (*.xlsx)",)

        if file != '':
            if 'xlsx' in filter_:
                f = file
                if not f.endswith('.xlsx'):
                    f += '.xlsx'
                self.logger.to_xlsx(f)

            if 'csv' in filter_:
                f = file
                if not f.endswith('.csv'):
                    f += '.csv'
                self.logger.to_csv(f)

    def copy_click(self):
        """
        Copy logs to the clipboard
        """
        df = self.logger.to_df()
        s = io.StringIO()
        df.to_csv(s, sep='\t')
        txt = s.getvalue()

        # copy to clipboard
        cb = QtWidgets.QApplication.clipboard()
        cb.clear(mode=cb.Clipboard)
        cb.setText(txt, mode=cb.Clipboard)


class ElementsDialogue(QtWidgets.QDialog):
    """
    Selected elements dialogue window
    """

    def __init__(self, name, elements: list()):
        super(ElementsDialogue, self).__init__()
        self.setObjectName("self")
        self.setContextMenuPolicy(QtCore.Qt.NoContextMenu)
        self.layout = QtWidgets.QVBoxLayout(self)

        # build elements list
        self.objects_table = QtWidgets.QTableView()

        if len(elements) > 0:
            model = ObjectsModel(elements, elements[0].editable_headers,
                                 parent=self.objects_table, editable=False,
                                 non_editable_attributes=[1, 2, 14])

            self.objects_table.setModel(model)

        # accept button
        self.accept_btn = QtWidgets.QPushButton()
        self.accept_btn.setText('Proceed')
        self.accept_btn.clicked.connect(self.accept_click)

        # Copy button
        self.copy_btn = QtWidgets.QPushButton()
        self.copy_btn.setText('Copy')
        self.copy_btn.clicked.connect(self.copy_click)

        # add all to the GUI
        self.layout.addWidget(self.objects_table)
        self.frame2 = QtWidgets.QFrame()
        self.layout.addWidget(self.frame2)
        self.layout2 = QtWidgets.QHBoxLayout(self.frame2)

        self.layout2.addWidget(self.accept_btn)
        # self.layout2.addWidget(QtWidgets.QSpacerItem())
        self.layout2.addWidget(self.copy_btn)

        self.setLayout(self.layout)

        self.setWindowTitle(name)

        self.accepted = False

    def accept_click(self):
        self.accepted = True
        self.accept()

    def copy_click(self):
        pass


class TimeReIndexDialogue(QtWidgets.QDialog):
    """
    New profile dialogue window
    """
    def __init__(self):
        super(TimeReIndexDialogue, self).__init__()
        self.setObjectName("self")
        self.setContextMenuPolicy(QtCore.Qt.NoContextMenu)
        self.main_layout = QtWidgets.QVBoxLayout(self)

        self.accepted = False

        self.label1 = QtWidgets.QLabel()
        self.label1.setText("Year")

        # year
        d = datetime.now()
        self.year_spinner = QtWidgets.QSpinBox()
        self.year_spinner.setMinimum(0)
        self.year_spinner.setMaximum(3000)
        self.year_spinner.setValue(d.year)

        self.label2 = QtWidgets.QLabel()
        self.label2.setText("Hours per interval")

        self.interval_hours = QtWidgets.QDoubleSpinBox()
        self.interval_hours.setMinimum(0.0001)
        self.interval_hours.setMaximum(1000)
        self.interval_hours.setValue(1)

        # accept button
        self.accept_btn = QtWidgets.QPushButton()
        self.accept_btn.setText('Accept')
        self.accept_btn.clicked.connect(self.accept_click)

        # add all to the GUI
        self.main_layout.addWidget(self.label1)
        self.main_layout.addWidget(self.year_spinner)
        self.main_layout.addWidget(self.label2)
        self.main_layout.addWidget(self.interval_hours)
        self.main_layout.addWidget(self.accept_btn)

        self.setLayout(self.main_layout)

        self.setWindowTitle('Time re-index')

        h = 120
        self.resize(h, int(1.1 * h))

    def accept_click(self):
        """
        Accept and close
        """
        self.accepted = True
        self.accept()


class CorrectInconsistenciesDialogue(QtWidgets.QDialog):
    """
    New profile dialogue window
    """
    def __init__(self):
        super(CorrectInconsistenciesDialogue, self).__init__()
        self.setObjectName("self")
        self.setContextMenuPolicy(QtCore.Qt.NoContextMenu)
        self.main_layout = QtWidgets.QVBoxLayout(self)

        self.accepted = False

        self.label1 = QtWidgets.QLabel()
        self.label1.setText("Minimum generator set point")

        # min voltage
        self.min_voltage = QtWidgets.QDoubleSpinBox()
        self.min_voltage.setMinimum(0)
        self.min_voltage.setMaximum(2)
        self.min_voltage.setSingleStep(0.01)
        self.min_voltage.setValue(0.98)

        self.label2 = QtWidgets.QLabel()
        self.label2.setText("Maximum generator set point")

        # min voltage
        self.max_voltage = QtWidgets.QDoubleSpinBox()
        self.max_voltage.setMinimum(0)
        self.max_voltage.setMaximum(2)
        self.max_voltage.setSingleStep(0.01)
        self.max_voltage.setValue(1.02)

        self.label3 = QtWidgets.QLabel()
        self.label3.setText("Maximum virtual tap difference")

        self.max_virtual_tap = QtWidgets.QDoubleSpinBox()
        self.max_virtual_tap.setMinimum(0)
        self.max_virtual_tap.setMaximum(1)
        self.max_virtual_tap.setSingleStep(0.01)
        self.max_virtual_tap.setValue(0.1)

        # accept button
        self.accept_btn = QtWidgets.QPushButton()
        self.accept_btn.setText('Accept')
        self.accept_btn.clicked.connect(self.accept_click)

        # add all to the GUI
        self.main_layout.addWidget(self.label1)
        self.main_layout.addWidget(self.min_voltage)
        self.main_layout.addWidget(self.label2)
        self.main_layout.addWidget(self.max_voltage)
        self.main_layout.addWidget(self.label3)
        self.main_layout.addWidget(self.max_virtual_tap)
        self.main_layout.addWidget(self.accept_btn)

        self.setLayout(self.main_layout)

        self.setWindowTitle('Correct inconsistencies')

        h = 120
        self.resize(h, int(1.1 * h))

    def accept_click(self):
        """
        Accept and close
        """
        self.accepted = True
        self.accept()


def clear_qt_layout(layout):
    """
    Remove all widgets from a layout object
    :param layout:
    """
    for i in reversed(range(layout.count())):
        layout.itemAt(i).widget().deleteLater()


if __name__ == "__main__":
    import sys
    from PySide2.QtWidgets import QApplication
    app = QApplication(sys.argv)
    window = LogsDialogue(name='', logger=Logger())
    window.show()
    sys.exit(app.exec_())



import sys
import os
from PySide2.QtWidgets import *
import folium
from shutil import copyfile
import webbrowser

try:
    from PySide2.QtWebEngineWidgets import QWebEngineView as QWebView, QWebEnginePage as QWebPage
    qt_web_engine_available = True
except ModuleNotFoundError:
    qt_web_engine_available = False
    print('QtWebEngineWidgets not found :(')

from GridCal.Gui.GIS.gui import *
from GridCal.Engine.IO.file_system import get_create_gridcal_folder


class GISWindow(QMainWindow):

    def __init__(self, external_file_path=''):
        """
        Constructor
        :param external_file_path: path to the file to open
        """
        QMainWindow.__init__(self)
        self.ui = Ui_GisWindow()
        self.ui.setupUi(self)
        self.setWindowTitle('GridCal - GIS')

        if os.path.exists(external_file_path):
            self.file_path = external_file_path
        else:
            self.file_path = self.generate_blank_map_html(lon_avg=40.430, lat_avg=3.56)

        # create web browser for the map
        if qt_web_engine_available:
            self.web_layout = QtWidgets.QVBoxLayout(self.ui.webFrame)
            self.webView = QWebView()
            self.web_layout.addWidget(self.webView)
            self.ui.webFrame.setContentsMargins(0, 0, 0, 0)
            self.web_layout.setContentsMargins(0, 0, 0, 0)
            self.webView.setUrl(QtCore.QUrl.fromLocalFile(self.file_path))
        else:
            webbrowser.open('file://' + self.file_path)

        # # action linking
        self.ui.actionSave_map.triggered.connect(self.save)

    def closeEvent(self, event):
        """
        Remove the file on close
        :param event:
        """
        if os.path.exists(self.file_path):
            os.remove(self.file_path)

    def save(self):
        """
        Save a copy of the displayed map
        :return:
        """
        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog

        file, filter = QFileDialog.getSaveFileName(self, "Save map", '',
                                                   filter="html (*.html)",
                                                   options=options)

        if file != '':
            if not file.endswith('.html'):
                file += '.html'

            copyfile(self.file_path, file)

    def msg(self, text, title="Warning"):
        """
        Message box
        :param text: Text to display
        :param title: Name of the window
        """
        msg = QMessageBox()
        msg.setIcon(QMessageBox.Information)
        msg.setText(text)
        # msg.setInformativeText("This is additional information")
        msg.setWindowTitle(title)
        # msg.setDetailedText("The details are as follows:")
        msg.setStandardButtons(QMessageBox.Ok)
        retval = msg.exec_()

    @staticmethod
    def generate_blank_map_html(lon_avg, lat_avg):
        """
        Generate a blank HTML map file
        :param lon_avg: center longitude
        :param lat_avg: center latitude
        :return: file name
        """
        my_map = folium.Map(location=[lon_avg, lat_avg], zoom_start=5)

        gc_path = get_create_gridcal_folder()

        path = os.path.join(gc_path, 'map.html')

        my_map.save(path)

        return path


if __name__ == "__main__":

    app = QtWidgets.QApplication(sys.argv)
    window = GISWindow()
    window.resize(1.61 * 700.0, 600.0)  # golden ratio
    window.show()
    sys.exit(app.exec_())

